from cli.config import login, initialize_config, switch_env, read_config
from cli.datasets import fetch_datasets, fetch_dataset, create_uploaded_dataset, create_generated_dataset
from cli.scenes import fetch_scenes, fetch_scene, create_scene
from cli.jobs import fetch_jobs, create_new_job
from cli.utils import to_pathlib_path, parse_args

import logging
import click
import json

log = logging.getLogger(__name__)


@click.group()
@click.option('--verbose', '-v', is_flag=True, help="Enables verbose mode.")
def cli(verbose=False):
    ''' zpy cli is client side ragnarok '''
    # Set up logging
    level = logging.INFO
    if verbose:
        level = logging.DEBUG
    logging.basicConfig(
        level=level, format='%(message)s')
    initialize_config()


@cli.command('help')
def _help():
    log.info('usage: zpy <command> [<args>]')


@cli.command('env')
@click.argument('env', type=click.Choice(['local', 'stage', 'prod']))
def _env(env):
    """ switch target environment """
    log.info(f'switching to environment {env}')
    switch_env(env)


@cli.command('login')
@click.argument('username')
@click.argument('password')
def _login(username, password):
    login(username, password)


@cli.command('config')
def _config():
    """ display config """
    config = read_config()
    log.info('zpy cli configuration:')
    log.info(config)


############
### LIST ###
############


@cli.group()
def list():
    """ list resources """
    pass


@list.command('datasets')
def list_datasets():
    config = read_config()
    fetch_datasets(config['ENDPOINT'], config['TOKEN'])


@list.command('scenes')
def list_scenes():
    config = read_config()
    fetch_scenes(config['ENDPOINT'], config['TOKEN'])


@list.command('jobs')
def list_jobs():
    config = read_config()
    fetch_jobs(config['ENDPOINT'], config['TOKEN'])


###########
### GET ###
###########


@cli.group()
def get():
    """ get resource """
    pass


@get.command('dataset')
@click.argument('name')
@click.argument('dtype', type=click.Choice(['job', 'generated', 'uploaded']))
@click.argument('path')
def get_dataset(name, dtype, path):
    config = read_config()
    dir_path = to_pathlib_path(path)
    if not dir_path.exists():
        log.info(f'output path {dir_path} does not exist')
        return
    fetch_dataset(name, path, dtype, config['ENDPOINT'], config['TOKEN'])


@get.command('scene')
@click.argument('name')
@click.argument('path')
def get_scene(name, path):
    config = read_config()
    dir_path = to_pathlib_path(path)
    if not dir_path.exists():
        log.info(f'output path {dir_path} does not exist')
        return
    fetch_scene(name, path, config['ENDPOINT'], config['TOKEN'])

##############
### UPLOAD ###
##############

@cli.group()
def upload():
    """ upload resource """
    pass


@upload.command('scene')
@click.argument('name')
@click.argument('path')
def upload_scene(name, path):
    config = read_config()
    input_path = to_pathlib_path(path)
    if not input_path.exists():
        log.warning(f'input path {input_path} does not exist')
        return
    if input_path.suffix != '.zip':
        log.warning(f'input path {input_path} not a zip file')
    create_scene(name, path, config['ENDPOINT'], config['TOKEN'])


@upload.command('dataset')
@click.argument('name')
@click.argument('path')
def upload_dataset(name, path):
    config = read_config()
    input_path = to_pathlib_path(path)
    if not input_path.exists():
        log.info(f'input path {input_path} does not exist')
        return
    if input_path.suffix != '.zip':
        log.warning(f'input path {input_path} not a zip file')
    create_uploaded_dataset(name, path, config['ENDPOINT'], config['TOKEN'])


##############
### CREATE ###
##############


@cli.group()
def create():
    """ create resource """
    pass


@create.command('dataset')
@click.argument('name')
@click.argument('scene')
@click.argument('args', nargs=-1)
def create_dataset(name, scene, args):
    config = read_config()
    dataset_config = parse_args(args)
    create_generated_dataset(name, scene, dataset_config, config['ENDPOINT'], config['TOKEN'])


@create.command('job')
@click.argument('name')
@click.argument('operation')
@click.option('datasets', '-d', multiple=True)
@click.argument('args', nargs=-1)
def create_job(name, operation, datasets, args):
    config = read_config()
    job_config = parse_args(args)
    datasets_list = [x for x in datasets]
    create_new_job(name, operation, job_config, datasets_list, config['ENDPOINT'], config['TOKEN'])



"""
    Base class for new annotation formats. Any new annotation format must be
    able to take a saver object and output annotations.
"""
from zpy.saver import Saver


class Output:
    """Holds the logic for outputting annotations to file."""

    def __init__(self, saver: Saver):
        """ Initialize from a Saver object. """
        self.saver = saver

    def output_annotations(self):
        """ Output annotation file. """
        raise NotImplementedError('Must implement output_annotations()')



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

import logging as py_logging
log = py_logging.getLogger(__name__)

# Import files first since it is used in other zpy modules
from zpy import files
# Color utilities will reset the random colors on import
from zpy import color
color.reset()
from zpy import gin
from zpy import image
from zpy import logging
from zpy import requests
# Output class (outputting annotations to file)
from zpy import output
from zpy import output_coco
from zpy import output_mot
from zpy import output_zumo
from zpy import output_csv
# Saver object (recording annotations at runtime)
from zpy import saver
from zpy import saver_image
from zpy import saver_video

# If your zpy library has an extra dependency
# which may or may not be installed on a user's
# system, make sure to wrap it in a try-catch
try:
    from zpy import blender
    from zpy import kdtree
    from zpy import keypoints
    from zpy import camera
    from zpy import material
    from zpy import objects
    from zpy import render
except ModuleNotFoundError:
    log.warn('Could not load blender utilities.')
try:
    from zpy import viz
except ModuleNotFoundError:
    log.warn('Could not load viz utilities.')



"""
    COCO (Common Objects in Context) dataset format.
"""
import copy
import logging
from pathlib import Path
from typing import List, Union

import gin

import zpy
from zpy.output import Output
from zpy.saver_image import ImageSaver

log = logging.getLogger(__name__)


class COCOParseError(Exception):
    """ Invalid COCO Annotation found when parsing data contents. """
    pass


@gin.configurable
class OutputCOCO(Output):
    """Holds the logic for outputting COCO annotations to file."""

    ANNOTATION_FILENAME = Path('coco_annotations.json')

    @gin.configurable
    def coco_info(self,
                  keys_to_add: List[str] = None
                  ):
        """ coco info """
        coco_info = {
            'description': self.saver.metadata['description'],
            'url': self.saver.metadata['url'],
            'version': self.saver.metadata['date_created'],
            'year': self.saver.metadata['year'],
            'contributor': self.saver.metadata['contributor'],
            'date_created': self.saver.metadata['date_created'],
            'save_path': self.saver.metadata['save_path'],
        }
        # Add any extra keys.
        if keys_to_add is not None:
            for key in keys_to_add:
                value = self.saver.metadata.get(key, None)
                if value is not None:
                    coco_info[key] = value
        return coco_info

    def coco_license(self):
        """ coco license """
        return {
            "url": "http://zumolabs.ai/image_license/",
            "id": 0,
            "name": "Zumo Labs Image License"
        }

    @gin.configurable
    def coco_categories(self,
                        keys_to_add: List[str] = [
                            'keypoints',
                            'skeleton',
                            'supercategory',
                            'subcategories',
                        ]
                        ):
        """ coco categories """
        coco_categories = []
        for category in self.saver.categories.values():
            coco_category = {
                'id': category['id'],
                'name': category['name'],
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = category.get(key, None)
                    if value is not None:
                        coco_category[key] = value
            coco_categories.append(coco_category)
        return coco_categories

    @gin.configurable
    def coco_images(self,
                    only_default_images: bool = True,
                    keys_to_add: List[str] = None,
                    ):
        """ coco images """
        _coco_img_id_to_idx = {}  # Image id does not correspond to index in image list
        coco_images = []
        for image in self.saver.images.values():
            if only_default_images and not image['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_img = {
                'license': 0,
                'id': image['id'],
                'file_name': image['name'],
                'coco_url': image['name'],
                'width': image['width'],
                'height': image['height'],
                'date_captured': self.saver.metadata['date_created'],
                'flickr_url': '.',
            }
            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = image.get(key, None)
                    if value is not None:
                        coco_img[key] = value
            coco_images.append(coco_img)
        return coco_images

    @gin.configurable
    def coco_annotations(self,
                         keys_to_add: List[str] = [
                             'bbox', 'area', 'segmentation'],
                         clipped: bool = True,
                         only_default_images: bool = True,
                         ):
        """ coco annotations """
        coco_annotations = []
        for annotation in self.saver.annotations:
            if only_default_images and \
                    not self.saver.images[annotation['image_id']]['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                'category_id': annotation['category_id'],
                'image_id': annotation['image_id'],
                'id': annotation['id'],
                'iscrowd': False,
            }
            if clipped:
                height = self.saver.images[annotation['image_id']]['height']
                width = self.saver.images[annotation['image_id']]['width']

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        if key == 'segmentation':
                            coco_ann['segmentation'] = self.saver.clip_coordinate_list(
                                width=width,
                                height=height,
                                annotation=annotation['segmentation']) if clipped \
                                else annotation['segmentation']
                        elif key == 'segmentation_rle':
                            coco_ann['segmentation_rle'] = annotation['segmentation_rle']
                        elif key == 'segmentation_float':
                            coco_ann['segmentation_float'] = self.saver.clip_coordinate_list(
                                normalized=True,
                                annotation=annotation['segmentation_float']) if clipped \
                                else annotation['segmentation_float']
                        elif key == 'bbox_float':
                            coco_ann['bbox_float'] = self.saver.clip_bbox(
                                normalized=True,
                                bbox=annotation['bbox_float']) if clipped \
                                else annotation['bbox_float']
                        elif key == 'bbox':
                            coco_ann['bbox'] = self.saver.clip_bbox(
                                width=width,
                                height=height,
                                bbox=annotation['bbox']) if clipped \
                                else annotation['bbox']
                        elif key == 'bboxes_float':
                            coco_ann['bboxes_float'] = [self.saver.clip_bbox(
                                normalized=True,
                                bbox=bbox) for
                                bbox in annotation['bboxes_float']] if clipped \
                                else annotation['bboxes_float']
                        elif key == 'bboxes':
                            coco_ann['bboxes'] = [self.saver.clip_bbox(
                                height=height,
                                width=width,
                                bbox=bbox) for
                                bbox in annotation['bboxes']] if clipped \
                                else annotation['bboxes']
                        else:
                            coco_ann[key] = value
                    try:
                        if key == 'area':
                            coco_ann['area'] = annotation['bbox'][2] * \
                                annotation['bbox'][3]
                        elif key == 'areas':
                            coco_ann['areas'] = [bbox[2] * bbox[3]
                                                 for bbox in annotation['bboxes']]
                    except:
                        pass
            # HACK: Require bbox for an annotation
            if coco_ann.get('bbox', None) is not None:
                coco_annotations.append(coco_ann)

        return coco_annotations

    @gin.configurable
    def coco_split_segmentation_annotations(self,
                                            keys_to_add: List[str] = [
                                                'bbox', 'area', 'segmentation'],
                                            clipped: bool = True,
                                            only_default_images: bool = True,
                                            ):
        """ coco annotations one per segmentation """
        coco_annotations = []
        # Annotation id will be re-mapped
        annotation_id = 0
        for annotation in self.saver.annotations:
            if only_default_images and \
                    not self.saver.images[annotation['image_id']]['style'] == 'default':
                # COCO annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            coco_ann = {
                'category_id': annotation['category_id'],
                'image_id': annotation['image_id'],
                'iscrowd': False,
            }
            if clipped:
                height = self.saver.images[annotation['image_id']]['height']
                width = self.saver.images[annotation['image_id']]['width']

            # Add any extra keys.
            if keys_to_add is not None:
                for key in keys_to_add:
                    value = annotation.get(key, None)
                    if value is not None:
                        coco_ann[key] = value

            # Annotations can be composed of multiple annotation components
            if annotation.get('segmentation') is not None:
                num_components = len(annotation['segmentation'])
            else:
                log.warning(
                    'Skipping annotation: split segmentation requires segmentaiton field.')
                continue

            # TODO: This can prolly be cleaned up?
            for i in range(num_components):
                _coco_ann = copy.deepcopy(coco_ann)
                try:
                    _coco_ann['segmentation'] = [self.saver.clip_coordinate_list(
                        height=height,
                        width=width,
                        annotation=annotation['segmentation'][i])] if clipped \
                        else [annotation['segmentation'][i]]
                except:
                    pass
                try:
                    _coco_ann['segmentation_rle'] = [
                        annotation['segmentation_rle'][i]]
                except:
                    pass
                try:
                    _coco_ann['segmentation_float'] = [self.saver.clip_coordinate_list(
                        normalized=True,
                        annotation=annotation['segmentation_float'][i])] if clipped \
                        else [annotation['segmentation_float'][i]]
                except:
                    pass
                try:
                    _coco_ann['bbox_float'] = self.saver.clip_bbox(
                        normalized=True,
                        bbox=annotation['bboxes_float'][i]) if clipped \
                        else annotation['bboxes_float'][i]
                except:
                    pass
                try:
                    _coco_ann['bbox'] = self.saver.clip_bbox(
                        width=width,
                        height=height,
                        bbox=annotation['bboxes'][i]) if clipped \
                        else annotation['bboxes'][i]
                except:
                    pass
                try:
                    _coco_ann['area'] = annotation['areas'][i]
                except:
                    pass

                # HACK: Require bbox for an annotation
                if _coco_ann.get('bbox', None) is not None:
                    _coco_ann['id'] = annotation_id
                    annotation_id += 1
                    coco_annotations.append(_coco_ann)

        return coco_annotations

    @gin.configurable
    def output_annotations(self,
                           splitseg: bool = False,
                           annotation_path: Union[str, Path] = None,
                           ):
        """ Output COCO annotations. """
        coco_dict = {
            'info': self.coco_info(),
            'licenses': self.coco_license(),
            'categories': self.coco_categories(),
            'images': self.coco_images(),
            'annotations': self.coco_annotations(),
        }

        # Get the correct annotation path
        if annotation_path is not None:
            annotation_path = annotation_path
        elif self.saver.annotation_path is None:
            annotation_path = self.saver.output_dir / self.ANNOTATION_FILENAME
        else:
            annotation_path = self.saver.annotation_path

        # Write out annotations to file
        zpy.files.write_json(annotation_path, coco_dict)
        parse_coco_annotations(
            annotation_file=annotation_path, data_dir=self.saver.output_dir)

        # Output split-segmentation annotations
        if splitseg:
            log.info('Outputting COCO annotations with multi-part' +
                     'segmentation split into seperate annotations')
            coco_dict['annotations'] = self.coco_split_segmentation_annotations()
            annotation_path = zpy.files.make_underscore_path(
                annotation_path, 'splitseg')

            # Write out annotations to file
            zpy.files.write_json(annotation_path, coco_dict)
            parse_coco_annotations(
                annotation_file=annotation_path, data_dir=self.saver.output_dir)


@gin.configurable
def parse_coco_annotations(
    annotation_file: Union[str, Path] = None,
    data_dir: Union[str, Path] = None,
    output_saver: bool = False,
    # Specify which keys to add to ImageSaver
    image_keys_to_add: List[str] = None,
) -> ImageSaver:
    """ Parse COCO annotations, optionally output a ImageSaver object. """
    log.info(f'Parsing COCO annotations at {annotation_file}...')
    # Check annotation file path
    annotation_file = zpy.files.verify_path(annotation_file)
    if data_dir is not None:
        data_dir = zpy.files.verify_path(data_dir, check_dir=True)
    else:
        # If no data_dir, assume annotation file is in the root folder.
        data_dir = annotation_file.parent

    # Check that categories, images, and annotations are not blank
    coco_annotations = zpy.files.read_json(annotation_file)
    images = coco_annotations["images"]
    if len(images) == 0:
        raise COCOParseError(f'no images found in {annotation_file}')
    categories = coco_annotations["categories"]
    if len(categories) == 0:
        raise COCOParseError(
            f'no categories found in {annotation_file}')
    annotations = coco_annotations["annotations"]
    if len(annotations) == 0:
        raise COCOParseError(
            f'no annotations found in {annotation_file}')
    log.info(
        f'images:{len(images)} categories:{len(categories)} annotations:{len(annotations)}')

    # Optionally output a saver object.
    if output_saver:
        saver = ImageSaver(output_dir=data_dir,
                           annotation_path=annotation_file,
                           description=coco_annotations['info']['description'],
                           clean_dir=False,
                           )

    # Check Images
    log.info('Parsing images...')
    img_ids = []
    for img in images:
        # Image ID
        image_id = img['id']
        if not isinstance(image_id, int):
            raise COCOParseError(f'image id {image_id} must be int.')
        if image_id in img_ids:
            raise COCOParseError(f'image id {image_id} already used.')
        img_ids.append(image_id)
        if image_id < 0:
            raise COCOParseError(f'invalid image id {image_id}')
        # Height and Width
        height, width = img['height'], img['width']
        if not isinstance(height, int):
            raise COCOParseError(f'height {height} must be int.')
        if not isinstance(width, int):
            raise COCOParseError(f'width {width} must be int.')
        if height <= 0 or width <= 0:
            raise COCOParseError(
                f'width and height h:{height} w:{width} must be > 0')
        # Image Name
        filename = img['file_name']
        if not isinstance(filename, str):
            raise COCOParseError(f'filename {filename} must be str.')
        image_path = data_dir / filename
        if not image_path.exists():
            raise COCOParseError(
                f'image path {image_path} does not exist')
        # COCO Path
        coco_url = img.get('coco_url', None)
        if coco_url is None:
            coco_url = filename
        coco_url = Path(coco_url)
        coco_path = data_dir / coco_url
        if not coco_path.exists():
            raise COCOParseError(f'coco url {coco_path} does not exist')
        # Save each image to ImageSaver object
        if output_saver:
            saver.images[image_id] = {
                'id': image_id,
                'name': filename,
                'output_path': str(coco_url),
                'height': height,
                'width': width,
                'style': 'default',
            }
            # Add any extra keys.
            if image_keys_to_add is not None:
                for key in image_keys_to_add:
                    value = img.get(key, None)
                    if value is not None:
                        saver.images[image_id][key] = value

    # Check Categories
    log.info('Parsing categories...')
    cat_ids = []
    cat_names = []
    for category in categories:
        name, category_id = category['name'], category['id']
        log.info(f'name:{name} id:{category_id}')
        # Category Name
        category_name = category['name']
        if not isinstance(category_name, str):
            raise COCOParseError(
                f'category_name {category_name} must be str.')
        if category_name in cat_names:
            raise COCOParseError(
                f'category_name {category_name} already used')
        cat_names.append(category_name)
        # Category ID
        category_id = category['id']
        if not isinstance(image_id, int):
            raise COCOParseError(
                f'category_id {category_id} must be int.')
        if category_id in cat_ids:
            raise COCOParseError(
                f'category id {category_id} already used')
        cat_ids.append(category_id)
        # Supercategories
        if category.get('supercategory', None) is not None:
            pass
        if category.get('supercategories', None) is not None:
            pass
        # Subcategories
        if category.get('subcategory', None) is not None:
            pass
        if category.get('subcategories', None) is not None:
            pass
        # Keypoints
        if category.get("keypoints", None) is not None:
            keypoints = category['keypoints']
            log.info(f"{len(keypoints)} keypoints:{keypoints}")
            if category.get('skeleton', None) is None:
                raise COCOParseError(
                    f'skeleton must be present with {keypoints}')
        # Save each category to ImageSaver object
        if output_saver:
            _category = saver.categories.get(category_id, None)
            if _category is None:
                saver.categories[category_id] = {}
            for key, value in category.items():
                saver.categories[category_id][key] = value

    # Check Annotations
    log.info('Parsing annotations...')
    ann_ids = []
    for annotation in annotations:
        # IDs
        image_id, category_id, annotation_id = annotation[
            'image_id'], annotation['category_id'], annotation['id']
        if image_id not in img_ids:
            raise COCOParseError(
                f'annotation img:{image_id} not in {img_ids}')
        if category_id not in cat_ids:
            raise COCOParseError(
                f'annotation cat:{category_id} not in {cat_ids}')
        if annotation_id in ann_ids:
            raise COCOParseError(
                f'annotation id:{annotation_id} already used')
        ann_ids.append(annotation_id)

        # Bounding Boxes
        bbox = annotation.get('bbox', None)
        if bbox is not None:
            pass

        # Keypoints
        keypoints = annotation.get('num_keypoints', None)
        if keypoints is not None:
            if "keypoints_xyv" in annotation:
                if len(annotation["keypoints_xyv"]) != int(annotation["num_keypoints"])*3:
                    raise COCOParseError(
                        'keypoints_xyv not correct size {len(keypoints)}')
            if "keypoints_xyz" in annotation:
                if len(annotation["keypoints_xyz"]) != int(annotation["num_keypoints"])*3:
                    raise COCOParseError(
                        'keypoints_xyz not correct size {len(keypoints)}')

        # Save each annotation to ImageSaver object
        if output_saver:
            saver.annotations.append(annotation)

    if output_saver:
        return saver
    else:
        return None



"""
    KDTree utilities for Blender Python.
"""
import logging
import random
from pathlib import Path
from typing import List, Tuple

import bpy
import gin
import mathutils
import numpy as np

import zpy

log = logging.getLogger(__name__)


def kdtree_from_collection(collections: List[bpy.types.Collection]) -> mathutils.kdtree.KDTree:
    """ Creates a KDTree of vertices from a collection of objects. """
    # First get the size of the objects (number of vertices)
    size = 0
    for obj in zpy.objects.for_obj_in_collections(collections):
        size += len(obj.data.vertices)
    # Then add them to a tree object
    kd = mathutils.kdtree.KDTree(size)
    insert_idx = 0
    for obj in zpy.objects.for_obj_in_collections(collections):
        for v in obj.data.vertices:
            world_coordinate_v = obj.matrix_world @ v.co
            kd.insert(world_coordinate_v, insert_idx)
            insert_idx += 1
    # Balancing is the most expensive operation
    kd.balance()
    return kd


@gin.configurable
def floor_occupancy(
    kdtree: mathutils.kdtree.KDTree,
    x_bounds: Tuple[float],
    y_bounds: Tuple[float],
    z_height: float = 0.0,
    num_voxels: int = 100,
) -> float:
    """ Get occupancy percentage for floor (XY plane). """
    log.info(f'Calculating floor occupancy ....')
    x_side_length = abs(x_bounds[1] - x_bounds[0])
    y_side_length = abs(y_bounds[1] - y_bounds[0])
    # Number of voxels determines number of points in each dimmension
    voxel_cube_side_length = (
        (x_side_length * y_side_length) / num_voxels)**(1/2)
    num_points_x = x_side_length / voxel_cube_side_length
    num_points_y = y_side_length / voxel_cube_side_length
    # TODO: This can definitely be vectorized better
    x_space, x_step = np.linspace(
        *x_bounds, num=int(num_points_x), retstep=True)
    y_space, y_step = np.linspace(
        *y_bounds, num=int(num_points_y), retstep=True)
    occupancy_grid = np.zeros((int(num_points_x), int(num_points_y)))
    for x_idx, x in enumerate(x_space):
        for y_idx, y in enumerate(y_space):
            x = float(x)
            y = float(y)
            closest_point = kdtree.find((x, y, z_height))[0]
            if (closest_point.x > (x - x_step)) and \
                    (closest_point.x < (x + x_step)):
                if (closest_point.y > (y - y_step)) and \
                        (closest_point.y < (y + y_step)):
                    occupancy_grid[x_idx][y_idx] = 1.0
    log.info(f'... Done.')
    log.debug(f'Floor occupancy grid: {str(occupancy_grid)}')
    return float(np.mean(occupancy_grid.copy()))


@gin.configurable
def volume_occupancy(
    kdtree: mathutils.kdtree.KDTree,
    x_bounds: Tuple[float],
    y_bounds: Tuple[float],
    z_bounds: Tuple[float],
    num_voxels: int = 100,
) -> float:
    """ Get occupancy percentage for volume. """
    log.info(f'Calculating volume occupancy ....')
    x_side_length = abs(x_bounds[1] - x_bounds[0])
    y_side_length = abs(y_bounds[1] - y_bounds[0])
    z_side_length = abs(z_bounds[1] - z_bounds[0])
    # Number of voxels determines number of points in each dimmension
    voxel_cube_side_length = (
        (x_side_length * y_side_length * z_side_length) / num_voxels)**(1/3)
    num_points_x = x_side_length / voxel_cube_side_length
    num_points_y = y_side_length / voxel_cube_side_length
    num_points_z = z_side_length / voxel_cube_side_length
    # TODO: This can definitely be vectorized better
    x_space, x_step = np.linspace(
        *x_bounds, num=int(num_points_x), retstep=True)
    y_space, y_step = np.linspace(
        *y_bounds, num=int(num_points_y), retstep=True)
    z_space, z_step = np.linspace(
        *z_bounds, num=int(num_points_z), retstep=True)
    occupancy_grid = np.zeros(
        (int(num_points_x), int(num_points_y), int(num_points_z)))
    for x_idx, x in enumerate(x_space):
        for y_idx, y in enumerate(y_space):
            for z_idx, z in enumerate(z_space):
                x = float(x)
                y = float(y)
                z = float(z)
                closest_point = kdtree.find((x, y, z))[0]
                if (closest_point.x > (x - x_step)) and \
                        (closest_point.x < (x + x_step)):
                    if (closest_point.y > (y - y_step)) and \
                            (closest_point.y < (y + y_step)):
                        if (closest_point.z > (z - z_step)) and \
                                (closest_point.z < (z + z_step)):
                            occupancy_grid[x_idx][y_idx][z_idx] = 1.0
    log.info(f'... Done.')
    log.debug(f'Volume occupancy grid: {str(occupancy_grid)}')
    return float(np.mean(occupancy_grid.copy()))



"""
    Utilities for multi-processing and requests.
"""
import json
import logging
import multiprocessing
import signal
import sys
import time
import traceback
from functools import partial, wraps
from pprint import pformat
from typing import Any, Dict

import zmq

import gin
import zpy

log = logging.getLogger(__name__)


class InvalidRequest(Exception):
    """ Network message to launcher is incorrect. """
    pass


def verify_key(
    request: Dict,
    key: str,
    key_type: type = None
) -> Any:
    """ Check request dict for key, raise error if not present, or if wrong type. """
    value = request.get(key, None)
    if value is None:
        raise InvalidRequest(f'Required key {key} not found.')
    if key_type is not None:
        if not isinstance(value, key_type):
            raise InvalidRequest(f'Key {key} must be of type {key_type}.')
    return value


class Process(multiprocessing.Process):
    """ Allows bubbiling up exceptions from a python process. """

    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            multiprocessing.Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((str(e), str(tb)))
            raise e

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


def request_as_process(request_func):
    """ Decorator for running a request as a separate process. """
    @wraps(request_func)
    def wrapped_request_func(request: Dict) -> None:
        _reply = multiprocessing.Manager().dict()
        p = Process(target=request_func, args=(request, _reply))
        p.start()
        p.join()
        global reply
        reply.update(_reply)
        if p.exception:
            reply['exception'] = p.exception[0]
            reply['trace'] = p.exception[1]
            reply['code'] = 400
    return wrapped_request_func


# Global signal variables (see func below)
abort = None
waiting = None
reply = None


def handle_signal(signum, frame) -> None:
    """ Handle interrupt signal. """
    log.info(f'Received interrupt signal {signum}')
    if waiting:
        sys.exit(1)
    global abort
    abort = True


def accept_requests(run_func):
    """ Decorator for accepting requests as seperate processes. """
    @wraps(run_func)
    def wrapped_run_func(bind_uri: str) -> None:
        """ Main entrypoint for request based communication. """
        log.info('Configuring zmq socket...')
        context = zmq.Context()
        socket = context.socket(zmq.REP)
        socket.bind(bind_uri)
        signal.signal(signal.SIGTERM, handle_signal)
        global abort, waiting, reply
        abort = False
        while not abort:
            log.info('Waiting for requests...')
            waiting = True
            request = json.loads(socket.recv_json())
            zpy.logging.linebreaker_log('new request')
            log.info(f'New request: {pformat(request)}')
            waiting = False

            # Reply will include duration of request
            start_time = time.time()
            try:
                # Request can set a log level
                log_level = request.get('log_level', None)
                if log_level is not None:
                    zpy.logging.set_log_levels(level=log_level)

                # Default reply will include a message and an error code
                reply = {
                    'request': request,
                    'code': 200,
                }

                # Reset any gin configs
                try:
                    gin.enter_interactive_mode()
                    gin.clear_config()
                except Exception as e:
                    log.warning(
                        f'Could not reset gin configs before request: {e}')

                # Call the function that was given
                run_func(request)

            except Exception as e:
                reply['exception'] = str(e)
                reply['code'] = 400

            # Duration of request is logged and sent in reply
            duration = time.time() - start_time
            reply['duration'] = duration

            # Send reply message back through the socket
            zpy.logging.linebreaker_log('reply')
            log.info(f'{pformat(reply)}')
            socket.send_json(json.dumps(reply))

        log.info('Exiting launcher.')

    return wrapped_run_func


def send_request(
    request: Dict,
    ip: str = '127.0.0.1',
    port: str = '5555',
) -> Dict:
    """ Send a request over a uri. """
    log.info(f'Connecting to {ip}:{port} ...')
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(f'tcp://{ip}:{port}')
    log.info('... Done!')
    log.info(f'Sending request: {request}')
    socket.send_json(json.dumps(request))
    log.info(f'Waiting for response...')
    response = json.loads(socket.recv_json())
    log.info(f'Received response: {pformat(response)}')
    return response



"""
    MOT (Multi Object Tracking) dataset format.
"""
import logging
from pathlib import Path
from typing import List, Union

import gin

import zpy
from zpy.output import Output
from zpy.saver_image import ImageSaver

log = logging.getLogger(__name__)


class MOTParseError(Exception):
    """ Invalid MOT Annotation found when parsing data contents. """
    pass


@gin.configurable
class OutputMOT(Output):
    """Holds the logic for outputting MOT annotations to file."""

    ANNOTATION_FILENAME = Path('mot.csv')

    @gin.configurable
    def output_annotations(self,
                           annotation_path: Union[str, Path] = None,
                           ):
        """ Ouput MOT (Multi-Object Tracking) annotations.

        https://motchallenge.net/faq/

        """
        mot = []
        for annotation in self.saver.annotations:
            if self.saver.images[annotation['image_id']]['style'] != 'default':
                # MOT annotations only have image annotations
                # for RGB images. No segmentation images.
                continue
            person_id = annotation.get('person_id', None)
            bbox = annotation.get('bbox', None)
            if (person_id is not None) and (bbox is not None):
                # Each CSV row will have 9 entries
                row = [0] * 9
                # Frame at which the object is present
                row[0] = annotation['frame_id']
                # Pedestrian trajectory is identiﬁed by a unique ID
                row[1] = person_id
                # Coordinate of the top-left corner of the pedestrian bounding box
                row[2] = bbox[0]
                # Coordinate of the top-left corner of the pedestrian bounding box
                row[3] = bbox[1]
                # Width in pixels of the pedestrian bounding box
                row[4] = bbox[2]
                # Height in pixels of the pedestrian bounding box
                row[5] = bbox[3]
                # Flag whether the entry is to be considered (1) or ignored (0).
                row[6] = 1
                # TODO: Type of object annotated
                '''
                MOT Types:                
                    Pedestrian 1
                    Person on vehicle 2
                    Car 3
                    Bicycle 4
                    Motorbike 5
                    Non motorized vehicle 6
                    Static person 7
                    Distractor 8
                    Occluder 9
                    Occluder on the ground 10
                    Occluder full 11
                    Reflection 12
                    Crowd 13
                '''
                row[7] = annotation['mot_type']
                # TODO: Visibility ratio, a number between 0 and 1 that says how much of that object
                # is visible. Can be due to occlusion and due to image border cropping.
                row[8] = 1.0
                # Add to mot list
                mot.append(row)
        # Get the correct annotation path
        if annotation_path is not None:
            annotation_path = annotation_path
        elif self.saver.annotation_path is None:
            annotation_path = self.saver.output_dir / self.ANNOTATION_FILENAME
        else:
            annotation_path = self.saver.annotation_path
        # Write out annotations to file
        zpy.files.write_csv(annotation_path, mot)
        # Verify annotations
        parse_mot_annotations(annotation_path)


@gin.configurable
def parse_mot_annotations(
    annotation_file: Union[str, Path],
) -> None:
    """ Parse MOT annotations """
    log.info(f'Verifying MOT annotations at {annotation_file}...')
    mot = zpy.files.read_csv(annotation_file)
    for row in mot:
        if not len(row) == 9:
            raise MOTParseError(
                f'Each row in MOT csv must have len 9, found len {len(row)}')
    pass



from cli.utils import auth_headers
from table_logger import TableLogger
import requests
import json
import logging

log = logging.getLogger(__name__)


def create_new_job(name, operation, config, datasets, url, token):
    """ create job on ragnarok """
    endpoint = f'{url}/api/v1/jobs/'
    data = {
        'operation': operation,
        'name': name,
        'input_data_sets': datasets,
        'config': json.dumps(config)
    }
    r = requests.post(endpoint, data=data, headers=auth_headers(token))
    if r.status_code != 201:
        log.warning(f'Unable to create {operation} job {name} on datasets {datasets}')
        return
    log.info(f'created {operation} job {name} {config} on datasets {datasets}')


def fetch_jobs(endpoint, token):
    """ fetch all datasets in ragnarok """
    endpoint = f'{endpoint}/api/v1/jobs/'
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning('Unable to fetch jobs')
        return
    jobs = json.loads(r.text)['results']
    tbl = TableLogger(columns='state,name,operation,created',default_colwidth=30)
    if len(jobs) == 0:
        log.info(None)
    for j in jobs:
        tbl(j['state'], j['name'], j['operation'], j['created_at'])



"""
    Base Saver object for storing annotations, categories, etc during scene runtime.
"""
import logging
from datetime import date
from pathlib import Path
from typing import Dict, List, Tuple, Union

import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class Saver:
    """Holds the logic for saving annotations at runtime."""

    # Names for annotation files, folders, configs, datasheets, etc
    HIDDEN_METAFOLDER_FILENAME = Path('.zumometa')
    HIDDEN_DATASHEET_FILENAME = Path('ZUMO_DATASHEET.txt')
    GIN_CONFIG_FILENAME = Path('config.gin')

    DATETIME_FORMAT = '20%y%m%d_%H%M%S'
    DATETIME_YEAR_FORMAT = '20%y'

    def __init__(self,
                 output_dir: Union[str, Path] = None,
                 annotation_path: Union[str, Path] = None,
                 description: str = 'Description of dataset.',
                 clean_dir: bool = True,
                 ):
        # the output dir
        if output_dir is None:
            output_dir = zpy.files.default_temp_path()
        self.output_dir = zpy.files.verify_path(
            output_dir, make=True, check_dir=True)
        log.debug(f'Saver output directory at {output_dir}')
        if clean_dir:
            zpy.files.clean_dir(self.output_dir)
        # Annotation files can be optionally written out to a different dir
        if annotation_path is None:
            self.annotation_path = annotation_path
        else:
            self.annotation_path = zpy.files.verify_path(annotation_path)
            log.debug(f'Saver annotation path at {annotation_path}')
        # Very similar keys to COCO-style
        self.metadata = {
            'description': description,
            'contributor': 'Zumo Labs',
            'url': 'zumolabs.ai',
            'year': date.today().strftime(self.DATETIME_YEAR_FORMAT),
            'date_created': date.today().strftime(self.DATETIME_FORMAT),
            'save_path': str(self.output_dir),
        }
        self.categories = {}
        self.annotations = []
        # Reverse-lookup dictionaries for name/color to id
        self.category_name_to_id = {}

    @gin.configurable
    def add_annotation(self,
                       category: str = None,
                       subcategory: str = None,
                       subcategory_zero_indexed: bool = True,
                       **kwargs,
                       ) -> Dict:
        """ Add annotation. """
        assert category is not None, 'Must provide a category for annotation.'
        category_id = self.category_name_to_id.get(category, None)
        assert category_id is not None, f'Could not find id for category {category}'
        self.categories[category_id]['count'] += 1
        if subcategory is not None:
            subcategory_id = self.categories[category_id]['subcategories'].index(
                subcategory)
            self.categories[category_id]['subcategory_count'][subcategory_id] += 1
            subcategory_id += 0 if subcategory_zero_indexed else 1
            annotation['subcategory_id'] = subcategory_id
        annotation['category_id'] = category_id
        annotation['id'] = len(self.annotations)
        return annotation

    @gin.configurable
    def add_category(self,
                     name: str = 'default',
                     supercategories: List[str] = None,
                     subcategories: List[str] = None,
                     color:  Tuple[float] = (0., 0., 0.),
                     zero_indexed: bool = True,
                     **kwargs,
                     ) -> Dict:
        """ Add category. """
        # Default for super- and sub- categories is empty list
        supercategories = supercategories or []
        subcategories = subcategories or []
        category = {
            'name': name,
            'supercategories': supercategories,
            'subcategories': subcategories,
            'color': color,
            'count': 0,
            'subcategory_count': [0] * len(subcategories),
        }
        category.update(**kwargs)
        category['id'] = len(self.categories.keys())
        category['id'] += 0 if zero_indexed else 1
        log.debug(f'Adding category: {zpy.files.pretty_print(category)}')
        self.categories[category['id']] = category
        self.category_name_to_id[name] = category['id']
        return category

    @gin.configurable
    def remap_filter_categories(self,
                                category_remap: Dict = None,
                                ) -> None:
        """ Remap and filter category ids and names. """
        if category_remap is None:
            log.warning(
                'Attempted to remap categories with no category remap.')
            return

        # Intermediate variables for organization
        category_remap_ids = []
        category_remap_name_to_id = {}
        category_remap_old_id_to_new_id = {}

        # Check for duplicates and typing
        for _id, _name in category_remap.items():
            try:
                _id = int(_id)
                _name = str(_name)
            except ValueError:
                raise ValueError('Category remap must be {int : string}')
            if _name in category_remap_name_to_id:
                raise ValueError(f'Duplicate category name in remap: {_name}')
            if _id in category_remap_ids:
                raise ValueError(f'Duplicate category id in remap: {_id}')
            category_remap_ids.append(_id)
            category_remap_name_to_id[_name] = _id

        # Make sure the category names all exist in current categories
        category_names = [c['name'] for c in self.categories.values()]
        for category_name in category_remap_name_to_id:
            assert category_name in category_names, \
                f'Could not find category {category_name} in dataset when remap-ing'

        # Go through all of the current categories
        new_categories = {}
        for old_id, category in self.categories.items():
            # Matching is done using the name
            if category['name'] in category_remap_name_to_id:
                new_id = category_remap_name_to_id[category['name']]
                category_remap_old_id_to_new_id[old_id] = new_id
                category['id'] = new_id
                new_categories[new_id] = category
        # Overwrite the old categories
        self.categories = new_categories

        # Go through all of the current annotations
        new_annotations = []
        # Replace the category_id in all annotations
        for annotation in self.annotations:
            if annotation['category_id'] in category_remap_old_id_to_new_id:
                new_id = category_remap_old_id_to_new_id[annotation['category_id']]
                annotation['category_id'] = new_id
                new_annotations.append(annotation)
        # Overwrite the old annotations
        self.annotations = new_annotations

    def output_gin_config(self):
        """ Output the full gin config. """
        gin_config_filepath = self.output_dir / self.GIN_CONFIG_FILENAME
        log.info(f'Writing out gin config to {gin_config_filepath}')
        with open(gin_config_filepath, "w") as f:
            f.write(gin.operative_config_str())

    @staticmethod
    def write_datasheet(
        datasheet_path: str = None,
        info: Dict = None
    ) -> None:
        with datasheet_path.open('w') as f:
            for k, v in info.items():
                f.write(f'{k},{v}\n')

    @staticmethod
    def clip_coordinate_list(
        annotation: List[Union[int, float]] = None,
        height: Union[int, float] = None,
        width: Union[int, float] = None,
        normalized: bool = False,
    ) -> List[Union[int, float]]:
        """ Clip a list of coordinates (e.g. segmentation polygon) """
        if any(isinstance(i, list) for i in annotation):
            return [zpy.saver.Saver.clip_coordinate_list(height=height,
                                                         width=width,
                                                         normalized=normalized,
                                                         annotation=ann) for ann in annotation]
        if normalized:
            # Coordinates are in (0, 1)
            max_x, max_y = 1.0, 1.0
        else:
            # Coordinates are w.r.t image height and width
            max_x, max_y = width, height
        new_annotation = []
        # TODO: This zip unpack here is unreadable
        for x, y in zip(*[iter(annotation)]*2):
            new_x, new_y = x, y
            if x < 0:
                new_x = 0
            if y < 0:
                new_y = 0
            if x > max_x:
                new_x = max_x
            if y > max_y:
                new_y = max_y
            new_annotation.append(new_x)
            new_annotation.append(new_y)
        return new_annotation

    @staticmethod
    def clip_bbox(
        bbox: List[Union[int, float]] = None,
        height: Union[int, float] = None,
        width: Union[int, float] = None,
        normalized: bool = False,
    ) -> List[Union[int, float]]:
        """ Clip a bounding box in [x, y, width, height] format. """
        if normalized:
            # Coordinates are in (0, 1)
            max_x, max_y = 1.0, 1.0
        else:
            # Coordinates are w.r.t image height and width
            max_x, max_y = width, height
        new_bbox = [0] * 4
        new_bbox[0] = max(0, min(bbox[0], max_x))
        new_bbox[1] = max(0, min(bbox[1], max_y))
        new_bbox[2] = max(0, min(bbox[2], (max_x - new_bbox[0])))
        new_bbox[3] = max(0, min(bbox[3], (max_y - new_bbox[1])))
        return new_bbox



from pathlib import Path
from typing import Union
import os
from tqdm import tqdm
from urllib.request import urlopen
from urllib.parse import urlparse
import logging
import random

log = logging.getLogger(__name__)


def _safe_eval(key):
    try:
        return eval(key)
    except:
        return key


def parse_args(args):
    keys = args[::2]
    vals = map(lambda x: _safe_eval(x), args[1::2])
    return dict(zip(keys,vals))


def download_url(url, output_path):
    u = urlopen(url)
    h = u.info()
    totalSize = int(h["Content-Length"])

    log.info(f"Downloading {totalSize} bytes...")
    fp = open(output_path, 'wb')

    blockSize = 8192
    with tqdm(total=totalSize) as pbar:
        while True:
            chunk = u.read(blockSize)
            if not chunk: 
                break
            fp.write(chunk)
            pbar.update(blockSize)

    fp.flush()
    fp.close()

    log.info('Downloaded:')
    log.info(f'url: {url}')
    log.info(f'local path: {output_path}')


def auth_headers(token):
    return {'Authorization': 'token {}'.format(token)}


def to_pathlib_path(path: Union[str, Path]) -> Path:
    """ Convert string path to pathlib.Path if needed. """
    if not isinstance(path, Path):
        path = Path(os.path.expandvars(path))
    return path



"""
    Segment panel and functions.
"""
import importlib
import json
import logging
import random
from typing import Tuple

import bpy
import zpy
from bpy.types import Operator
from bpy_extras.io_utils import ImportHelper

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerObjectProperties():
    """ Properties applied to object."""
    bpy.types.Object.seg = bpy.props.PointerProperty(
        type=SegmentableProperties)


def registerSceneProperties():
    """ Properties applied to scenes."""
    bpy.types.Scene.categories = bpy.props.CollectionProperty(
        type=CategoryProperties
    )
    bpy.types.Scene.categories_enum = bpy.props.EnumProperty(
        name="Category",
        description="Category for this object.",
        default=None,
        items=_category_items,
        update=_category_update,
    )


class CategoryProperties(bpy.types.PropertyGroup):
    """ Segmentation category is a property of one or many objects. """
    name: bpy.props.StringProperty(
        name="Category Name",
        description="String name of the category.",
    )
    color: bpy.props.FloatVectorProperty(
        name="Category Color",
        subtype='COLOR',
        description="Category color for segmentation.",
    )


def _category_items(self, context):
    """ Get current categories. """
    _categories_enum = []
    for i, (name, category) in enumerate(bpy.context.scene.categories.items()):
        # Add category to ENUM tuple
        _categories_enum.append((
            # First item is used for Python access.
            str(i),
            name,
            zpy.color.frgb_to_hex(category.color),
        ))
    return _categories_enum


def _category_update(self, context):
    """ Update the category. """
    if context.selected_objects:
        # Use the value of the category enum dropdown
        category = context.scene.categories[int(context.scene.categories_enum)]
        for obj in zpy.objects.for_obj_in_selected_objs(context):
            zpy.objects.segment(
                obj=obj,
                name=category.name,
                color=category.color,
                as_category=True,
            )


class SegmentableProperties(bpy.types.PropertyGroup):
    category_name: bpy.props.StringProperty(
        name="Category Name",
        description="String name of the category.",
        default='',
    )
    category_color: bpy.props.FloatVectorProperty(
        name="Category Color",
        subtype='COLOR',
        description="Category color for segmentation.",
    )
    instance_name: bpy.props.StringProperty(
        name="Instance Name",
        description="String name of the instance.",
        default='',
    )
    instance_color: bpy.props.FloatVectorProperty(
        name="Instance Color",
        subtype='COLOR',
        description="Instance color for segmentation.",
    )


class SegmentInstanceMany(Operator):
    """ Segment the selected objects/parts.

    Each object will be segmented as a unique object.

    """
    bl_idname = "object.zpy_segment_instance_many"
    bl_label = "Segment Instance (Many)"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return len(context.selected_objects) > 0

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'
        for obj in zpy.objects.for_obj_in_selected_objs(context):
            zpy.objects.segment(obj=obj, name=obj.name)
        return {'FINISHED'}


class SegmentInstanceSingle(Operator):
    """ Segment the selected objects/parts.

    All objects will be segmented as a single instance.

    """
    bl_idname = "object.zpy_segment_instance_single"
    bl_label = "Segment Instance (Single)"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return len(context.selected_objects) > 0

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'
        # Pick a random color and instance name
        _name = context.selected_objects[0].name
        _color = zpy.color.random_color(output_style='frgb')
        for obj in zpy.objects.for_obj_in_selected_objs(context):
            zpy.objects.segment(obj=obj, name=_name, color=_color)
        return {'FINISHED'}


class VisualizeInstance(Operator):
    """ Visualize the instance colors on objects in scene. """
    bl_idname = "object.zpy_visualize_instance"
    bl_label = "Visualize Instances"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'
        # Loop through all objects in the scene
        for obj in context.scene.objects:
            if not obj.type == 'MESH':
                continue
            context.view_layer.objects.active = obj
            if obj.seg.instance_color is not None:
                obj.color = zpy.color.frgb_to_frgba(
                    obj.seg.instance_color)
            else:
                obj.color = zpy.color.default_color(output_style='frgba')
        return {'FINISHED'}


class VisualizeCategory(Operator):
    """ Visualize the category colors on objects in scene. """
    bl_idname = "object.zpy_visualize_category"
    bl_label = "Visualize Categories"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'
        # Loop through all objects in the scene
        for obj in context.scene.objects:
            if not obj.type == 'MESH':
                continue
            context.view_layer.objects.active = obj
            if obj.seg.category_color is not None:
                obj.color = zpy.color.frgb_to_frgba(
                    obj.seg.category_color)
            else:
                obj.color = zpy.color.default_color()
        return {'FINISHED'}


class ResetSegData(Operator):
    """ Reset the segmentation data on the selected objects/parts. """
    bl_idname = "object.zpy_reset_seg_data"
    bl_label = "Reset Segmentation Data"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return len(context.selected_objects) > 0

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'
        for obj in context.selected_objects:
            # Only meshes or empty objects TODO: Why the empty objects
            if not (obj.type == 'MESH' or obj.type == 'EMPTY'):
                continue
            obj.seg.instance_name = ''
            obj.seg.instance_color = zpy.color.default_color(
                output_style='frgb')
            obj.seg.category_name = 'default'
            obj.seg.category_color = zpy.color.default_color(
                output_style='frgb')
            obj.color = zpy.color.default_color(output_style='frgba')
        return {'FINISHED'}


def _reset_categories(context):
    """ Reset the scene categories. """
    # Clean out the scene-level category dict
    for _ in range(len(context.scene.categories)):
        context.scene.categories.remove(0)
    # Reset all categories
    for obj in zpy.objects.for_obj_in_selected_objs(context):
        obj.seg.category_name = 'default'
        obj.seg.category_color = zpy.color.default_color(
            output_style='frgb')
        obj.color = zpy.color.default_color(output_style='frgba')


def _add_category(context,
                  name: str = None,
                  color: Tuple[float] = None) -> None:
    """ Add category to enum category property. """
    if name in context.scene.categories.keys():
        log.warning(f'Skipping duplicate category {name}.')
        return
    if color is None:
        color = zpy.color.random_color(output_style='frgb')
        log.info(f'Choosing random color for category {name}: {color}')
    # Add category to categories dict
    new_category = context.scene.categories.add()
    new_category.name = name
    new_category.color = color


class CategoriesFromText(Operator):
    """ Populate categories from text block. """
    bl_idname = "object.zpy_categories_from_text"
    bl_label = "Categories from Text"

    def execute(self, context):

        # BUG: Clicking "Text" resets all the categories and their colors

        txt = bpy.data.texts
        if "categories" in txt.keys():
            category_text = txt["categories"]

        else:
            txt.new("categories")
            category_text = txt["categories"]

        assert category_text is not None, \
            f'Category text block must exist for segmentation.'

        # Activate the categories text block in the text editor
        for area in context.screen.areas:
            if area.type == 'TEXT_EDITOR':
                space = area.spaces.active
                space.text = category_text

        _reset_categories(context)

        for line in category_text.lines:
            _add_category(context, name=line.body)
        return {'FINISHED'}


class CategoriesFromZUMOJSON(Operator, ImportHelper):
    """ Populate categories from Zumo JSON. """
    bl_idname = "object.zpy_categories_from_zumo"
    bl_description = "Categories from Zumo JSON"
    bl_label = "Import"

    filter_glob: bpy.props.StringProperty(
        default='*.json', options={'HIDDEN'})

    def execute(self, context):
        zumo_json = zpy.files.read_json(self.filepath)
        categories = zumo_json.get('categories', None)
        assert categories is not None, \
            f'ZUMO JSON does not have categories.'
        _reset_categories(context)
        for category in categories.values():
            _add_category(
                context,
                name=category.get('name', None),
                color=category.get('color', None),
            )
        return {'FINISHED'}


class SegmentPanel(bpy.types.Panel):
    """ UI for the addon that is visible in Blender. """
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Segment"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout

        row = layout.row()
        row.label(text="Segment Selected Objects")
        row = layout.row()
        row.operator(
            'object.zpy_segment_instance_single',
            text='Single',
            icon='USER',
        )
        row.operator(
            'object.zpy_segment_instance_many',
            text='Many',
            icon='COMMUNITY',
        )
        row = layout.row()
        row.prop(context.scene, "categories_enum", text="")
        row = layout.row()
        row.label(text="Visualize")
        row = layout.row()
        row.operator(
            'object.zpy_visualize_instance',
            text='Visualize Instances',
            icon='HIDE_OFF',
        )
        row = layout.row()
        row.operator(
            'object.zpy_visualize_category',
            text='Visualize Categories',
            icon='HIDE_OFF',
        )

        row = layout.row()
        row.label(text="Load Categories")
        row = layout.row()
        row.operator(
            'object.zpy_categories_from_text',
            text='Text',
            icon='TEXT',
        )
        row.operator(
            'object.zpy_categories_from_zumo',
            text='Json',
            icon='FILEBROWSER',
        )

        row = layout.row()
        row.label(text="Selected Object Data")
        row = layout.row()
        row.operator(
            'object.zpy_reset_seg_data',
            text='Reset Seg Data',
            icon='FILE_REFRESH',
        )
        row = layout.row()
        row.label(text="Instance")
        row = layout.row()
        row.prop(context.object.seg, "instance_name", text="")
        row.prop(context.object.seg, "instance_color", text="")
        row = layout.row()
        row.label(text="Category")
        row = layout.row()
        row.prop(context.object.seg, "category_name", text="")
        row.prop(context.object.seg, "category_color", text="")
        row = layout.row()



"""
    Blender utilities.
"""
import inspect
import logging
import math
import random
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union

import bpy
import gin
import mathutils
import numpy as np

import zpy

log = logging.getLogger(__name__)


def use_gpu() -> None:
    """ Use GPU for rendering. """
    devices = list(
        bpy.context.preferences.addons['cycles'].preferences.devices)
    log.debug(f'Devices available {devices}')
    prefs = bpy.context.preferences.addons['cycles'].preferences
    prefs.compute_device_type = 'CUDA'
    devices = prefs.get_devices()
    for device in devices[0]:
        device.use = True
    log.debug(f'Devices available {devices}')


@gin.configurable
def set_seed(seed: int = 0) -> None:
    """ Set the random seed. """
    log.info(f'Setting random seed to {seed}')
    if log.getEffectiveLevel() == logging.DEBUG:
        # When debugging you want to run into errors related
        # to specific permutations of the random variables, so
        # you need to vary the seed to run into them.
        seed = random.randint(1, 100)
        log.debug(f'Choosing a random random seed of {seed}')
    random.seed(seed)
    np.random.seed(seed)


def set_log_levels(level: str = None) -> None:
    """ Set logger levels for all zpy modules. """
    if level is None:
        log_level = logging.INFO
    elif level == 'info':
        log_level = logging.INFO
    elif level == 'debug':
        log_level = logging.DEBUG
    elif level == 'warning':
        log_level = logging.WARNING
    else:
        log.warning(f'Invalid log level {level}')
        return
    log.warning(f'Setting log level to {log_level}')
    for logger_name in ['zpy', 'zpy_addon', 'zpy.savers', 'bpy.zpy_addon']:
        logging.getLogger(logger_name).setLevel(log_level)


@gin.configurable
def step(
    num_steps: int = 16,
    framerate: int = 0,
    start_frame: int = 1,
    refresh_ui: bool = False,
) -> int:
    """ Step logic helper for the scene. """
    assert num_steps is not None, 'Invalid num_steps'
    assert num_steps > 0, 'Invalid num_steps'
    step_idx = 0
    if framerate > 0:
        start = bpy.context.scene.frame_start
        stop = bpy.context.scene.frame_end
        log.info(f'Animation enabled. Min frames: {start}. Max frames: {stop}')
    while step_idx < num_steps:
        log.info('-----------------------------------------')
        log.info('                   STEP                  ')
        log.info('-----------------------------------------')
        log.info(f'Simulation step {step_idx} of {num_steps}.')
        start_time = time.time()
        if framerate > 0:
            current_frame = start_frame + step_idx * framerate
            bpy.context.scene.frame_set(current_frame)
            log.info(f'Animation frame {bpy.context.scene.frame_current}')
        # # Update the step_idx for all RandomEvent and Animator instances
        # RandomEvent.step_idx = step_idx
        yield step_idx
        step_idx += 1
        duration = time.time() - start_time
        log.info(f'Simulation step took {duration}s to complete.')
        # TODO: This call is not needed in headless instances, makes loop faster
        if refresh_ui:
            refresh_blender_ui()


def connect_debugger_vscode(timeout: int = 3) -> None:
    """ Connects to a VSCode debugger.

    Based on:

    https://github.com/AlansCodeLog/blender-debugger-for-vscode

    """
    if log.getEffectiveLevel() == logging.DEBUG:
        log.debug('Starting VSCode debugger in Blender.')
        # TODO: Can we assume the user will properly set up this environment variable?
        path = '$BLENDERADDONS/blender-debugger-for-vscode/__init__.py'
        path = zpy.files.verify_path(path, make=False)
        bpy.ops.preferences.addon_install(filepath=str(path))
        bpy.ops.preferences.addon_enable(module='blender-debugger-for-vscode')
        bpy.ops.debug.connect_debugger_vscode()
        for sec in range(timeout):
            log.debug(f'You have {timeout - sec} seconds to connect!')
            time.sleep(1)


def parse_config(text_name: str = 'config') -> None:
    """ Load gin config for scene """
    _text = bpy.data.texts.get(text_name, None)
    if _text is None:
        log.warning(f'Could not find {text_name} in texts.')
        return
    log.info(f'Loading gin config {text_name}')
    gin.enter_interactive_mode()
    with gin.unlock_config():
        gin.parse_config(_text.as_string())
        gin.finalize()


def run_text(text_name: str = 'run') -> None:
    """ Run a text script in Blender. """
    _text = bpy.data.texts.get(text_name, None)
    if _text is None:
        log.warning(f'Could not find {text_name} in texts.')
        return
    _ctx = bpy.context.copy()
    _ctx['edit_text'] = _text
    bpy.ops.text.run_script(_ctx)


def connect_addon(name: str = 'zpy_addon') -> None:
    """ Connects a Blender AddOn. """
    log.debug(f'Connecting Addon {name}.')
    path = f'$BLENDERADDONS/{name}/__init__.py'
    path = zpy.files.verify_path(path, make=False)
    bpy.ops.preferences.addon_install(filepath=str(path))
    bpy.ops.preferences.addon_enable(module=name)


def output_intermediate_scene(path: Union[str, Path] = None) -> None:
    """ Output intermediate saved scene. """
    if path is None:
        path = zpy.files.default_temp_path() / 'blender-debug-scene-tmp.blend'
    path = zpy.files.verify_path(path, make=False)
    log.debug(f'Saving intermediate scene to {path}')
    bpy.ops.wm.save_as_mainfile(filepath=str(path), compress=False, copy=True)


def refresh_blender_ui() -> None:
    """ Refresh blender in the middle of a script.

    Does not work on headless instances.
    """
    log.debug(f'Refreshing Blender UI.')
    bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
    bpy.context.view_layer.update()


def load_scene(
    path: Union[str, Path],
    auto_execute_scripts : bool = True,
    ) -> None:
    """ Load a scene from a *.blend file. """
    # HACK: Clear out scene of cameras and lights
    clear_scene(['CAMERA', 'LIGHT'])
    path = zpy.files.verify_path(path, make=False)
    log.debug(f'Loading scene from {str(path)}.')
    with bpy.data.libraries.load(str(path)) as (data_from, data_to):
        for attr in dir(data_to):
            setattr(data_to, attr, getattr(data_from, attr))
    # HACK: Delete current empty scene
    bpy.ops.scene.delete()
    # HACK: Delete extra workspaces that are created e.g. 'Animation.001'
    _workspaces = [ws for ws in bpy.data.workspaces if '.0' in ws.name]
    bpy.data.batch_remove(ids=_workspaces)
    # Allow execution of scripts inside loaded scene
    if auto_execute_scripts:
        log.warning('Allowing .blend file to run scripts automatically')
        log.warning('   this is unsafe for blend files from an untrusted source')
        bpy.context.preferences.filepaths.use_scripts_auto_execute = auto_execute_scripts


def clear_scene(to_clear: List = ["MESH"]) -> None:
    """ Empty out the scene. """
    log.debug('Deleting all mesh objects in the scene.')
    for obj in bpy.data.objects:
        if obj.type in to_clear:
            bpy.data.objects.remove(obj)


def load_text_from_file(
    path: Union[str, Path],
    text_name: str = '',
) -> None:
    """ Load a file into Blender's internal text UI. """
    path = zpy.files.verify_path(path)
    if bpy.data.texts.get(text_name, None) is None:
        _text = bpy.data.texts.load(str(path), internal=True)
        _text.name = text_name
    else:
        bpy.data.texts[text_name].from_string(path.read_text())


def scene_information() -> Dict:
    """ Get the run() function kwargs. """
    log.info(f'Collecting scene information')
    run_script = bpy.data.texts.get('run', None)
    if run_script is None:
        raise ValueError('No run script found in scene.')
    # HACK: Gin is confused by the as_module() call
    gin.enter_interactive_mode()
    run_script_module = bpy.data.texts['run'].as_module()
    scene_doc = inspect.getdoc(run_script_module)

    run_function = None
    for name, value in inspect.getmembers(run_script_module):
        if name == 'run':
            run_function = value
    if run_function is None:
        raise ValueError('No run() function found in run script.')
    if not inspect.isfunction(run_function):
        raise ValueError('run() is not a function in run script.')

    run_kwargs = []
    for param in inspect.signature(run_function).parameters.values():
        _kwarg = {}
        _kwarg['name'] = param.name
        _kwarg['type'] = str(param.annotation)
        _kwarg['default'] = param.default
        run_kwargs.append(_kwarg)

    _ = {
        'name': bpy.context.scene.zpy_scene_name,
        'version': bpy.context.scene.zpy_scene_version,
        'description': scene_doc,
        'run_kwargs': run_kwargs,
        'export_date':  time.strftime("%m%d%Y_%H%M_%S"),
        'zpy_version': zpy.__version__,
        'zpy_path': zpy.__file__,
        'blender_version': '.'.join([str(_) for _ in bpy.app.version]),
    }
    log.info(f'{_}')
    return _


@gin.configurable
def load_hdri(
    path: Union[str, Path],
    scale: Tuple[float] = (1.0, 1.0, 1.0),
) -> None:
    """ Load an HDRI from path.

    Great source of HDRIs:

        https://hdrihaven.com/

    """
    log.info(f'Loading HDRI at {path}')
    path = zpy.files.verify_path(path, make=False)
    world = bpy.context.scene.world
    world.use_nodes = True
    out_node = world.node_tree.nodes.get('World Output', None)
    if out_node is None:
        out_node = world.node_tree.nodes.new('World Output')
    bg_node = world.node_tree.nodes.get('Background', None)
    if bg_node is None:
        bg_node = world.node_tree.nodes.new('Background')
    env_node = world.node_tree.nodes.get('Environment Texture', None)
    if env_node is None:
        env_node = world.node_tree.nodes.new('ShaderNodeTexEnvironment')
    env_node.image = bpy.data.images.load(str(path))
    env_node.texture_mapping.scale = mathutils.Vector(scale)
    # env_node.texture_mapping.rotation = mathutils.Vector(
    #     (0,-math.radians(90), math.radians(180)))
    # world.node_tree.links.new(env_node.inputs[0], tex_node.outputs[4])
    world.node_tree.links.new(bg_node.inputs[0], env_node.outputs[0])
    world.node_tree.links.new(out_node.inputs[0], bg_node.outputs[0])


@gin.configurable
def random_hdri(
    asset_dir: Union[str, Path] = '$ASSETS/lib/hdris/hdri_maker_lib/04k_Library',
) -> Path:
    """ Generate a random HDRI from an asset path. """
    asset_directory = zpy.files.verify_path(
        asset_dir, make=False, check_dir=True)
    hdri_path = random.choice([x for x in asset_directory.iterdir() if x.is_file()])
    load_hdri(hdri_path)
    return hdri_path



"""
    Utilities for Materials in Blender.
"""
import logging
import random
from pathlib import Path
from typing import Tuple, Union

import bpy
import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
def make_aov_material_output_node(
    mat: bpy.types.Material = None,
    obj: bpy.types.Object = None,
    style: str = 'instance',
) -> None:
    """ Make AOV Output nodes in Composition Graph. """
    # Make sure engine is set to Cycles
    if not (bpy.context.scene.render.engine == "CYCLES"):
        log.warning(' Setting render engine to CYCLES to use AOV')
        bpy.context.scene.render.engine == "CYCLES"
    # Only certain styles are available
    valid_styles = ['instance', 'category']
    assert style in valid_styles, \
        f'Invalid style {style} for AOV material output node, must be in {valid_styles}.'

    # HACK: multiple material slots
    all_mats = []

    # Use material
    if mat is not None:
        all_mats = [mat]
    # Get material from object
    elif obj is not None:
        if obj.active_material is None:
            log.debug(f'No active material found for {obj.name}')
            return
        if len(obj.material_slots) > 1:
            for mat in obj.material_slots:
                all_mats.append(mat.material)
        else:
            all_mats.append(obj.active_material)
    else:
        raise ValueError('Must pass in an Object or Material')

    # HACK: multiple material slots
    for mat in all_mats:

        # Make sure material is using nodes
        if not mat.use_nodes:
            mat.use_nodes = True
        _tree = mat.node_tree

        # Vertex Color Node
        _name = f'{style} Vertex Color'
        vertexcolor_node = _tree.nodes.get(_name)
        if vertexcolor_node is None:
            vertexcolor_node = _tree.nodes.new('ShaderNodeVertexColor')
        vertexcolor_node.layer_name = style
        vertexcolor_node.name = _name

        # AOV Output Node
        _name = style
        # HACK: property "name" of ShaderNodeOutputAOV behaves strangely with .get()
        aovoutput_node = None
        for _node in _tree.nodes:
            if _node.name == _name:
                aovoutput_node = _node
        if aovoutput_node is None:
            aovoutput_node = _tree.nodes.new('ShaderNodeOutputAOV')
        aovoutput_node.name = style
        _tree.links.new(vertexcolor_node.outputs['Color'],
                        aovoutput_node.inputs['Color'])


@gin.configurable
def jitter(
    mat: bpy.types.Material = None,
    roughness_std: float = 0.2,
    metallic_std: float = 0.2,
    specular_std: float = 0.2,
) -> None:
    "Randomize a real texture a little."

    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    if bsdf_node is None:
        log.warning(f'No BSDF node to jitter for material {mat.name}')
        return
    bsdf_node.inputs['Roughness'].default_value += random.gauss(
        0, roughness_std)
    bsdf_node.inputs['Metallic'].default_value += random.gauss(
        0, metallic_std)
    bsdf_node.inputs['Specular'].default_value += random.gauss(
        0, specular_std)


@gin.configurable
def make_mat(
    name: str = None,
    texture_path: Union[str, Path] = None,
    color: Tuple[float] = None,
) -> bpy.types.Material:
    """ Makes a material from a texture or color."""
    mat = bpy.data.materials.get(name)
    if mat is not None:
        log.debug(f'Material {name} already exists')
        return mat
    mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    out_node = mat.node_tree.nodes.get('Material Output')
    if color is not None:
        mat.node_tree.nodes.remove(bsdf_node)
        bsdf_node = mat.node_tree.nodes.new('ShaderNodeBsdfDiffuse')
        bsdf_node.inputs['Color'].default_value = color + (1.,)
        mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    elif texture_path is not None:
        assert texture_path is not None, 'Must provide texture path.'
        tex_node = mat.node_tree.nodes.new('ShaderNodeTexImage')
        tex_node.name = 'ImageTexture'
        coord_node = mat.node_tree.nodes.new('ShaderNodeTexCoord')
        texture_path = zpy.files.verify_path(texture_path, make=False)
        bpy.ops.image.open(filepath=str(texture_path))
        tex_node.image = bpy.data.images[texture_path.name]
        tex_node.image.colorspace_settings.name = 'Filmic Log'
        mat.node_tree.links.new(tex_node.outputs[0], bsdf_node.inputs[0])
        mat.node_tree.links.new(coord_node.outputs[0], tex_node.inputs[0])
        tex_node.image.reload()
    else:
        raise ValueError('make_mat requries either color or texture path.')
    return mat


def set_mat(
    obj: Union[str, bpy.types.Object] = None,
    mat: Union[str, bpy.types.Material] = None,
    recursive: bool = True,
) -> None:
    """ Recursively sets object material.

    Allows string material and object names as input.
    """
    if obj is None:
        log.warning('Empty object.')
        return
    if mat is None:
        log.warning('Empty material.')
        return
    if type(obj) == str:
        obj = bpy.data.objects.get(obj)
        if obj is None:
            log.warning(f'Could not find object {obj}.')
            return
    if type(mat) == str:
        mat = bpy.data.materials.get(mat)
        if mat is None:
            log.warning(f'Could not find material {mat}.')
            return
    if hasattr(obj, 'active_material'):
        log.debug(f'Setting object {obj.name} material {mat.name}')
        obj.active_material = mat
    else:
        log.warning('Object does not have material property')
        return
    # Recursively change material on all children of object
    if recursive:
        for child in obj.children:
            set_mat(child, mat)



"""
    Export panel and functions.
"""
import importlib
import logging
import time
from pathlib import Path

import bpy
import zpy

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """ Properties applied to scenes."""
    bpy.types.Scene.zpy_scene_name = bpy.props.StringProperty(
        name="Scene Name",
        description="Name of the scene, must match data portal.",
        default="default",
        # set=set_scene_name,
    )
    bpy.types.Scene.zpy_scene_version = bpy.props.StringProperty(
        name="Scene Version",
        description="Version of the scene, must match data portal.",
        default="0",
        # set=set_scene_version,
    )
    bpy.types.Scene.zpy_export_path = bpy.props.StringProperty(
        name='Export Path',
        description="Export path for packaged zumo scenes.",
        default=str(zpy.files.default_temp_path()),
        subtype='DIR_PATH',
    )


def set_scene_name(self, value) -> None:
    """ Fix the scene name: all lowercase maximum 10 letters """
    self.zpy_scene_name = value.lower()[:10]


def set_scene_version(self, value) -> None:
    """ Fix the scene version: integer > 0 """
    try:
        self['zpy_scene_version'] = abs(int(value))
    except ValueError:
        log.warning('Scene version must be a int')
        self.zpy_scene_version = '0'


class OpenExportDirOperator(bpy.types.Operator):
    """ Open file browser at export dir. """
    bl_idname = "scene.zpy_open_export_dir"
    bl_label = "Open Export Dir"
    bl_description = "Open file browser at export dir."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        zpy.files.open_folder_in_explorer(
            context.scene.zpy_export_path, make=True)
        return {'FINISHED'}


class CleanUpDirOperator(bpy.types.Operator):
    """ Clean up and package scene into export dir. """
    bl_idname = "scene.zpy_cleanup_scene"
    bl_label = "Clean Up Export Dir"
    bl_description = "Clean up export dir."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        log.info('Cleaning up scene.')
        # Remove any backup blender files
        zpy.files.remove_files_with_suffix(
            path=context.scene.zpy_export_path,
            exts=['.blend1', '.blend2', '.blend3'],
        )
        # Remove any existing nfo files
        zpy.files.remove_files_with_suffix(
            path=context.scene.zpy_export_path,
            exts=['.nfo'],
        )
        # TODO: Scene based clean up collections and objects listings (in the text editor)
        # TODO: Remove the custom scene scripts that are not needed for staging (keep run, config, categories for now)
        return {'FINISHED'}


class ExportOperator(bpy.types.Operator):
    """ Export scene for ingest to Data Portal. """
    bl_idname = "scene.zpy_export_scene"
    bl_label = "Export scene"
    bl_description = "Export scene for ingest to Data Portal."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        # Shows progress cursor in Blender UI.
        log.info('Export Started.')
        bpy.context.window_manager.progress_begin(0, 100)

        # Clean scene before every export
        bpy.ops.scene.zpy_cleanup_scene()

        # Create export directory in the Blender filepath
        export_dir_name = f'{context.scene.zpy_scene_name}_v{context.scene.zpy_scene_version}'
        export_path = Path(context.scene.zpy_export_path) / export_dir_name
        zpy.files.verify_path(export_path, make=True)

        # Find missing files before export
        log.info('Export Step 1 of 4: Checking for any missing files.')
        bpy.context.window_manager.progress_update(10)
        _path = zpy.files.verify_path('$ASSETS', make=False)
        bpy.ops.file.find_missing_files(directory=str(_path))

        # Fix all the asset paths by packing them into the .blend
        # file and then un-packing them into a 'textures' folder.
        try:
            bpy.ops.wm.save_as_mainfile(
                filepath=str(export_path / 'main.blend'),
                compress=False,
            )
            log.info('Export Step 2 of 4: Packing files into .blend.')
            bpy.context.window_manager.progress_update(30)
            bpy.ops.file.make_paths_absolute()
            bpy.ops.file.make_paths_relative()
            bpy.ops.file.pack_all()
            bpy.ops.file.unpack_all(method='WRITE_LOCAL')
            bpy.ops.wm.save_as_mainfile(
                filepath=str(export_path / 'main.blend'),
                compress=False,
            )
        except Exception as e:
            self.report({'ERROR'}, str(e))
            log.warning(f'Exception when exporting: {e}')
            return {'CANCELLED'}

        log.info('Export Step 3 of 4: Saving meta-information.')
        bpy.context.window_manager.progress_update(70)

        # Save nfo file with some meta information
        save_time = time.strftime("%m%d%Y_%H%M_%S")
        nfo_file = export_path / \
            f'v{context.scene.zpy_scene_version}_{save_time}.nfo'
        nfo_file.write_text('')

        # Output scene information in ZUMO_META
        zpy.files.write_json(
            export_path / 'ZUMO_META.json',
            zpy.blender.scene_information(),
        )

        # TODO: Export glTF into zip directory

        # Zip up the exported directory for easy upload
        log.info('Export Step 4 of 4: Zipping up package.')
        bpy.context.window_manager.progress_update(90)
        zpy.files.zip_file(
            in_path=export_path,
            zip_path=Path(context.scene.zpy_export_path) /
            f'{export_dir_name}.zip',
        )

        # Clean scene after every export
        bpy.ops.scene.zpy_cleanup_scene()

        log.info('Export Completed.')
        bpy.context.window_manager.progress_end()
        return {'FINISHED'}


class ExportPanel(bpy.types.Panel):
    """ UI for the addon that is visible in Blender. """
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Export"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row()
        row.operator(
            'scene.zpy_export_scene',
            text='Export Scene',
            icon='EXPORT',
        )
        row = layout.row()
        row.prop(scene, "zpy_scene_name", text="Name")
        row = layout.row()
        row.prop(scene, "zpy_scene_version", text="Version")
        row = layout.row()
        row.label(text="Export Path")
        row = layout.row()
        row.prop(scene, "zpy_export_path", text="")
        row = layout.row()
        row.operator(
            'scene.zpy_open_export_dir',
            text='Open',
            icon='FILEBROWSER',
        )
        row.operator(
            'scene.zpy_cleanup_scene',
            text='Clean',
            icon='PACKAGE',
        )



"""
    Rendering panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy
from bpy.types import Operator

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """ Properties applied to scenes."""
    bpy.types.Scene.zpy_output_path = bpy.props.StringProperty(
        name='Output Path',
        description="Output path for rendered images, annotations, etc.",
        default=str(zpy.files.default_temp_path()),
        subtype='DIR_PATH',
    )


class RenderOperator(Operator):
    """ Render out single image (rgb, segmented, depth). """
    bl_idname = "scene.zpy_render"
    bl_label = "Render"
    bl_description = "Render out segmented images."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        # TODO: Make sure scene is good to render?
        return True

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'

        # Image names
        rgb_image_name = zpy.files.make_rgb_image_name(0)
        cseg_image_name = zpy.files.make_cseg_image_name(0)
        iseg_image_name = zpy.files.make_iseg_image_name(0)
        depth_image_name = zpy.files.make_depth_image_name(0)

        # Output path
        output_path = Path(context.scene.zpy_output_path)

        # Save renders to file
        zpy.render.render_aov(
            rgb_path=output_path / rgb_image_name,
            iseg_path=output_path / iseg_image_name,
            cseg_path=output_path / cseg_image_name,
            depth_path=output_path / depth_image_name,
            width=640,
            height=480,
        )

        return {'FINISHED'}


class CleanOutputDirOperator(bpy.types.Operator):
    """ Clean up output dir. """
    bl_idname = "scene.zpy_cleanup_output_dir"
    bl_label = "Clean Output Dir"
    bl_description = "Clean output dir."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        zpy.files.clean_dir(context.scene.zpy_output_path, keep_dir=True)
        return {'FINISHED'}


class OpenOutputDirOperator(Operator):
    """ Open file browser at output dir. """
    bl_idname = "scene.zpy_open_output_dir"
    bl_label = "Open Output Dir"
    bl_description = "Open file browser at output dir."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        zpy.files.open_folder_in_explorer(
            context.scene.zpy_output_path, make=True)
        return {'FINISHED'}


class RenderPanel(bpy.types.Panel):
    """ UI for the addon that is visible in Blender. """
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Render"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.operator(
            'scene.zpy_render',
            text='Render',
            icon='FILE_IMAGE',
        )
        row = layout.row()
        row.label(text="Output Path")
        row = layout.row()
        row.prop(scene, "zpy_output_path", text="")
        row = layout.row()
        row.operator(
            'scene.zpy_open_output_dir',
            text='Open',
            icon='FILEBROWSER',
        )
        row.operator(
            'scene.zpy_cleanup_output_dir',
            text='Clean',
            icon='PACKAGE',
        )



"""
    Image utilties.
"""
import logging
import random
from itertools import groupby
from pathlib import Path
from typing import Dict, List, Union

import gin
import numpy as np
from scipy import ndimage as ndi
from shapely.geometry import MultiPolygon, Polygon
from skimage import color, exposure, img_as_uint, io, measure
from skimage.exposure import match_histograms
from skimage.morphology import binary_closing, binary_opening
from skimage.transform import resize

import zpy

log = logging.getLogger(__name__)


def open_image(image_path: Union[str, Path]) -> np.ndarray:
    """Open image from path to ndarray."""
    image_path = zpy.files.verify_path(image_path, make=False)
    img = None
    try:
        img = io.imread(image_path)
        if img.shape[2] > 3:
            log.debug('RGBA image detected!')
            img = img[:, :, :3]
        if img.max() > 2.0:
            img = np.divide(img, 255.0)
    except Exception as e:
        log.error(f'Error {e} when opening {image_path}')
    return img


def remove_alpha_channel(
    image_path: Union[str, Path]
) -> None:
    """Remove the alpha channel in an image."""
    img = open_image(image_path)
    io.imsave(image_path, img)
    log.info(f'Saving image with no alpha channel at {image_path}')


@gin.configurable
def jpeg_compression(
    image_path: Union[str, Path],
    quality : int = 40,    
) -> Path:
    """Add jpeg compression to an image."""
    image_path = zpy.files.verify_path(image_path, make=False)
    img = io.imread(image_path)
    # Make sure image is jpeg
    if not image_path.suffix == '.jpeg':
        image_path = image_path.with_suffix('.jpeg')
    io.imsave(image_path, arr=img, quality=quality)
    log.info(f'Saving compressed image at {image_path}')
    return image_path


@gin.configurable
def resize_image(
    image_path: Union[str, Path],
    width: int = 640,
    height: int = 480,
) -> None:
    """Change the size of the image."""
    img = open_image(image_path)
    resized_img = resize(img, (height, width), anti_aliasing=True)
    io.imsave(image_path, resized_img)


def pixel_mean_std(flat_images: List[np.ndarray]) -> Dict:
    """ Return the pixel mean and std (as floats and in 256 mode from a flattened images array. """
    # HACK: Incorrect type assumption
    flat_images = flat_images[0]
    if np.amax(flat_images) > 1:
        std_256 = np.std(flat_images, axis=0)
        mean_256 = np.mean(flat_images, axis=0)
        std = std_256 / 256
        mean = mean_256 / 256
    else:
        std = np.std(flat_images, axis=0)
        mean = np.mean(flat_images, axis=0)
        std_256 = std * 256.0
        mean_256 = mean * 256.0
    return {
        "mean": mean,
        "std": std,
        "mean_256": mean_256,
        "std_256": std_256,
    }


def flatten_images(
    images: List[np.ndarray],
    max_pixels: int = 500000,
) -> List[np.ndarray]:
    """ Flatten out images in a list. """
    flat_images = []
    for image in images:
        dims = np.shape(image)
        if len(dims) == 3:
            flat_images.append(np.reshape(
                image, (dims[0] * dims[1], dims[2])))
    flat_images = np.concatenate(flat_images, axis=0)
    subsample = flat_images[np.random.randint(flat_images.shape[0],
                                              size=max_pixels), :]
    return [subsample]


def pad_with(vector, pad_width, iaxis, kwargs):
    """
        https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    """
    pad_value = kwargs.get('padder', 10)
    vector[:pad_width[0]] = pad_value
    vector[-pad_width[1]:] = pad_value


def binary_mask_to_rle(binary_mask) -> Dict:
    """ Converts a binary mask to a RLE (run-length-encoded) dictionary.

    From:
    https://stackoverflow.com/questions/49494337/encode-numpy-array-using-uncompressed-rle-for-coco-dataset

    """
    binary_mask = np.asfortranarray(binary_mask)
    rle = {'counts': [], 'size': list(binary_mask.shape)}
    counts = rle.get('counts')
    for i, (value, elements) in enumerate(groupby(binary_mask.ravel(order='F'))):
        if i == 0 and value == 1:
            counts.append(0)
        counts.append(len(list(elements)))
    return rle


@gin.configurable
def seg_to_annotations(
        image_path: Path,
        remove_salt: bool = True,
        rle_segmentations: bool = False,
        float_annotations: bool = False,
        max_categories: int = 1000):
    """ Convert a segmentation image and bounding box to polygon segmentations. """
    log.info(f'Extracting annotations from segmentation: {image_path}')
    image_path = zpy.files.verify_path(image_path, make=False)
    img = open_image(image_path)
    img_height, img_width = img.shape[0], img.shape[1]
    # Unique colors represent each unique category
    unique_colors = np.unique(img.reshape(-1, img.shape[2]), axis=0)
    # Store bboxes, seg polygons, and area in annotations list
    annotations = []
    # Loop through each category
    if unique_colors.shape[0] > max_categories:
        raise ValueError(
            f'Over {max_categories} categories: {unique_colors.shape[0]}')
    for i in range(unique_colors.shape[0]):
        seg_color = unique_colors[i, :]
        log.debug(f'Unique color {seg_color}')
        if all(np.equal(seg_color, np.zeros(3))):
            log.debug('Color is background.')
            continue
        # Make an image mask for this category
        masked_image = img.copy()
        mask = (img != seg_color).any(axis=-1)
        masked_image[mask] = np.zeros(3)
        masked_image = color.rgb2gray(masked_image)
        if log.getEffectiveLevel() == logging.DEBUG:
            masked_image_name = str(
                image_path.stem) + f'_masked_{i}' + str(image_path.suffix)
            masked_image_path = image_path.parent / masked_image_name
            io.imsave(masked_image_path,
                      img_as_uint(exposure.rescale_intensity(masked_image)))
        if remove_salt:
            # Remove "salt"
            # https://scikit-image.org/docs/dev/api/skimage.morphology
            masked_image = binary_opening(masked_image)
            masked_image = binary_closing(masked_image)
        else:
            masked_image = binary_opening(masked_image)
        # HACK: Pad masked image so segmented objects that extend beyond
        #       image are properly contoured
        masked_image = np.pad(masked_image, 1, pad_with, padder=False)
        # RLE encoded segmentation from binary image
        if rle_segmentations:
            rle_segmentation = binary_mask_to_rle(masked_image)
        # Fill in the holes
        filled_masked_image = ndi.binary_fill_holes(masked_image)
        # Get countours for each blob
        contours = measure.find_contours(
            filled_masked_image, 0.01, positive_orientation='low')
        log.debug(
            f'found {len(contours)} contours for {seg_color} in {image_path}')
        # HACK: Sometimes all you get is salt for an image, in this case
        #       do not add any annotation for this category.
        if len(contours) == 0:
            continue
        segmentations = []
        segmentations_float = []
        bboxes = []
        bboxes_float = []
        areas = []
        areas_float = []
        polygons = []
        for contour in contours:
            # Flip from (row, col) representation to (x, y)
            # and subtract the padding pixel
            for j in range(len(contour)):
                row, col = contour[j]
                contour[j] = (col - 1, row - 1)
            # Make a polygon and simplify it
            poly = Polygon(contour)
            poly = poly.simplify(1.0, preserve_topology=True)
            polygons.append(poly)
            # Segmentation
            segmentation = np.array(poly.exterior.coords).ravel().tolist()
            segmentations.append(segmentation)
            segmentations_float.append([
                x/img_height if k % 2 == 0 else x/img_width for k, x in enumerate(segmentation)
            ])
            # Bounding boxes
            x, y, max_x, max_y = poly.bounds
            bbox = (x, y, max_x - x, max_y - y)
            bbox_float = [
                bbox[0] / img_width,
                bbox[1] / img_height,
                bbox[2] / img_width,
                bbox[3] / img_height,
            ]
            bboxes.append(bbox)
            bboxes_float.append(bbox_float)
            # Areas
            areas.append(poly.area)
            areas_float.append(poly.area / (img_width * img_height))
        # Combine the polygons to calculate the bounding box and area
        multi_poly = MultiPolygon(polygons)
        x, y, max_x, max_y = multi_poly.bounds
        bbox = (x, y, max_x - x, max_y - y)
        area = multi_poly.area
        bbox_float = [
            bbox[0] / img_width,
            bbox[1] / img_height,
            bbox[2] / img_width,
            bbox[3] / img_height,
        ]
        area_float = area / (img_width * img_height)
        annotation = {
            'color': tuple(seg_color),
            # COCO standards
            'segmentation': segmentations,
            'bbox': bbox,
            'area': area,
            # List of list versions
            'bboxes': bboxes,
            'areas': areas,
        }
        if rle_segmentations:
            annotation['segmentation_rle'] = rle_segmentation
        if float_annotations:
            annotation['segmentation_float'] = segmentations_float
            annotation['bbox_float'] = bbox_float
            annotation['area_float'] = area_float
            annotation['bboxes_float'] = bboxes_float
            annotation['areas_float'] = areas_float
        annotations.append(annotation)
    return annotations


"""
    Human Keypoint Skeleton Formats.
"""
import logging

import bpy
import gin

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class Keypoints:
    """ Functionality for Keypoints."""

    # COCO(From pycoco)
    COCO_NAMES = [
        "nose",
        "left_eye", "right_eye",
        "left_ear", "right_ear",
        "left_shoulder", "right_shoulder",
        "left_elbow", "right_elbow",
        "left_wrist", "right_wrist",
        "left_hip", "right_hip",
        "left_knee", "right_knee",
        "left_ankle", "right_ankle"
    ]
    COCO_CONNECTIVITY = [[16, 14], [14, 12], [17, 15], [15, 13],
                         [12, 13], [6, 12], [7, 13], [6, 7],
                         [6, 8], [7, 9], [8, 10], [9, 11],
                         [2, 3], [1, 2], [1, 3], [2, 4],
                         [3, 5], [4, 6], [5, 7]]
    COCO_BONE_LOOKUP_MIXAMO = {
        "nose": "NoseEnd",
        "left_eye": "LeftEyeEnd",
        "right_eye": "RightEyeEnd",
        "left_ear": "LeftEarEnd",
        "right_ear": "RightEarEnd",
        "left_shoulder": "LeftArm",
        "right_shoulder": "RightArm",
        "left_elbow": "LeftForeArm",
        "right_elbow": "RightForeArm",
        "left_wrist": "LeftHand",
        "right_wrist": "RightHand",
        "left_hip": "LeftUpLeg",
        "right_hip": "RightUpLeg",
        "left_knee": "LeftLeg",
        "right_knee": "RightLeg",
        "left_ankle": "LeftFoot",
        "right_ankle": "RightFoot",
    }
    COCO_BONE_LOOKUP_ANIMA = {
        # TODO: Anima armature does not have eyes, ears, nose
        "nose": "Head",
        "left_eye": "Head",
        "right_eye": "Head",
        "left_ear": "Head",
        "right_ear": "Head",
        "left_shoulder": "LeftArm",
        "right_shoulder": "RightArm",
        "left_elbow": "LeftForeArm",
        "right_elbow": "RightForeArm",
        "left_wrist": "LeftHand",
        "right_wrist": "RightHand",
        "left_hip": "LeftUpLeg",
        "right_hip": "RightUpLeg",
        "left_knee": "LeftLeg",
        "right_knee": "RightLeg",
        "left_ankle": "LeftFoot",
        "right_ankle": "RightFoot",
    }

    # Body25B
    BODY25B_NAMES = [
        "Nose", "Neck",
        "RShoulder", "RElbow", "RWrist",
        "LShoulder", "LElbow", "LWrist",
        "RHip", "RKnee", "RAnkle",
        "LHip", "LKnee", "LAnkle",
        "REye", "LEye", "REar", "LEar",
    ]
    BODY25B_CONNECTIVITY = [[0, 1], [0, 2], [1, 3], [2, 4], [5, 7], [6, 8],
                            [7, 9], [8, 10], [5, 11], [6, 12], [11, 13],
                            [12, 14], [13, 15], [14, 16], [15, 19], [19, 20],
                            [15, 21], [16, 22], [22, 23], [16, 24],
                            [5, 17], [6, 17], [17, 18], [11, 12]]
    BODY25B_BONE_LOOKUP_MIXAMO = {
        "Nose": "NoseEnd",
        "Neck": "Head",
        "RShoulder": "RightShoulder",
        "RElbow": "RightForeArm",
        "RWrist": "RightHand",
        "LShoulder": "LeftShoulder",
        "LElbow": "LeftForeArm",
        "LWrist": "LeftHand",
        "RHip": "RightUpLeg",
        "RKnee": "RightLeg",
        "RAnkle": "RightFoot",
        "LHip": "LeftUpLeg",
        "LKnee": "LeftLeg",
        "LAnkle": "LeftFoot",
        "REye": "RightEyeEnd",
        "LEye": "LeftEyeEnd",
        "REar": "RightEarEnd",
        "LEar": "LeftEarEnd",
    }
    # TODO: Anima armature
    BODY25B_BONE_LOOKUP_ANIMA = {
        "Nose": "Head",
        "Neck": "Head",
        "RShoulder": "RightShoulder",
        "RElbow": "RightForeArm",
        "RWrist": "RightHand",
        "LShoulder": "LeftShoulder",
        "LElbow": "LeftForeArm",
        "LWrist": "LeftHand",
        "RHip": "RightUpLeg",
        "RKnee": "RightLeg",
        "RAnkle": "RightFoot",
        "LHip": "LeftUpLeg",
        "LKnee": "LeftLeg",
        "LAnkle": "LeftFoot",
        "REye": "RightEyeEnd",
        "LEye": "LeftEyeEnd",
        "REar": "RightEarEnd",
        "LEar": "LeftEarEnd",
    }

    def __init__(self,
                 root: bpy.types.Object,
                 style: str = 'coco',
                 armature: str = 'anima',
                 ):
        """ Initialize keypoint object. """
        if style == 'coco':
            self.names = self.COCO_NAMES
            self.connectivity = self.COCO_CONNECTIVITY
            if armature == 'mixamo':
                self.bone_lookup = self.COCO_BONE_LOOKUP_MIXAMO
            elif armature == 'anima':
                self.bone_lookup = self.COCO_BONE_LOOKUP_ANIMA
            else:
                raise ValueError(f'Unknown keypoint armature: {armature}')
        elif style == 'body25b':
            self.names = self.BODY25B_NAMES
            self.connectivity = self.BODY25B_CONNECTIVITY
            if armature == 'mixamo':
                self.bone_lookup = self.BODY25B_BONE_LOOKUP_MIXAMO
            elif armature == 'anima':
                self.bone_lookup = self.BODY25B_BONE_LOOKUP_ANIMA
            else:
                raise ValueError(f'Unknown keypoint armature: {armature}')
        else:
            raise ValueError(f'Unknown keypoint style: {style}')
        self.root = root
        self.style = style
        self.armature = armature
        self.bones = {bone.name: bone for bone in self.root.pose.bones}
        self.num_keypoints = None
        self.keypoints_xyv = None
        self.keypoints_xyz = None

    def update(self,
               world_transform=None,
               ) -> None:
        """ Add a keypoint skeleton. """
        self.num_keypoints = 0
        self.keypoints_xyv = []
        self.keypoints_xyz = []
        for name, bone_name in self.bone_lookup.items():
            bone = self.bones.get(bone_name, None)
            if bone is None:
                log.warning(
                    f'Could not find keypoint bone {name} using {bone_name}')
            if world_transform is None:
                pos = self.root.matrix_world @ bone.head
            else:
                pos = world_transform @ self.root.matrix_world @ bone.head
            x, y, v = zpy.camera.camera_xyv(pos, obj=self.root)
            self.keypoints_xyv += [x, y, v]
            self.keypoints_xyz += tuple(pos)
            self.num_keypoints += 1



"""
    Script loading/running panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy
from bpy.types import Operator

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


def registerSceneProperties():
    """ Properties applied to scenes."""
    bpy.types.Scene.zpy_gin_config_path = bpy.props.StringProperty(
        name='',
        description="Path to a gin config file.",
        default='',
        subtype='FILE_PATH',
        update=_load_gin_config,
    )
    bpy.types.Scene.zpy_runpy_path = bpy.props.StringProperty(
        name='',
        description="Path to run.py file",
        default='',
        subtype='FILE_PATH',
        update=_load_runpy,
    )


def _load_gin_config(self, context) -> None:
    """ Load gin config from file. """
    bpy.ops.scene.zpy_load_gin_config()


def _load_runpy(self, context) -> None:
    """ Load run.py from file. """
    bpy.ops.scene.zpy_load_runpy()


class RunOperator(Operator):
    """ Launch the run script in Blender's texts. """
    bl_idname = "scene.zpy_run"
    bl_label = "Run"
    bl_description = "Launch the run script in Blender's texts."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        # Save the state of the scene before the run script was executed
        bpy.ops.wm.save_mainfile()
        try:
            zpy.blender.use_gpu()
            zpy.blender.parse_config(LoadGinConfigOperator.DEFAULT_TEXT_NAME)
            zpy.blender.run_text(LoadRunpyOperator.DEFAULT_TEXT_NAME)
        except Exception as e:
            log.error(f'Executing script failed with exception {e}')
        # Return to the state of the scene before the run script was executed
        bpy.ops.wm.revert_mainfile()
        return {'FINISHED'}


class LoadGinConfigOperator(bpy.types.Operator):
    """ Load gin config from file. """
    bl_idname = "scene.zpy_load_gin_config"
    bl_label = "Load gin config from file."
    bl_description = "Load gin config from file."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    # Default name of the texts in Blender when loading
    DEFAULT_TEXT_NAME = "config"

    def execute(self, context):
        zpy.blender.load_text_from_file(
            bpy.path.abspath(context.scene.zpy_gin_config_path),
            text_name=self.DEFAULT_TEXT_NAME)
        return {'FINISHED'}


class PushGinConfigOperator(bpy.types.Operator):
    """ Push gin config to file. """
    bl_idname = "scene.zpy_push_gin_config"
    bl_label = "Push gin config to file."
    bl_description = "Push gin config to file."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        _text = bpy.data.texts[LoadGinConfigOperator.DEFAULT_TEXT_NAME].as_string(
        )
        with open(bpy.path.abspath(context.scene.zpy_gin_config_path), 'w') as _file:
            _file.write(_text)
        return {'FINISHED'}


class LoadRunpyOperator(bpy.types.Operator):
    """ Load run.py from file. """
    bl_idname = "scene.zpy_load_runpy"
    bl_label = "Load run.py from file."
    bl_description = "Load run.py from file."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    # Default name of the texts in Blender when loading
    DEFAULT_TEXT_NAME = "run"

    def execute(self, context):
        zpy.blender.load_text_from_file(
            bpy.path.abspath(context.scene.zpy_runpy_path),
            text_name=self.DEFAULT_TEXT_NAME)
        return {'FINISHED'}


class PushRunpyOperator(bpy.types.Operator):
    """ Push run.py to file. """
    bl_idname = "scene.zpy_push_runpy"
    bl_label = "Push run.py to file."
    bl_description = "Push run.py to file."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        _text = bpy.data.texts[LoadRunpyOperator.DEFAULT_TEXT_NAME].as_string()
        with open(bpy.path.abspath(context.scene.zpy_runpy_path), 'w') as _file:
            _file.write(_text)
        return {'FINISHED'}


class ScriptPanel(bpy.types.Panel):
    """ UI for the addon that is visible in Blender. """
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Script"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.operator(
            'scene.zpy_run',
            text='Run',
            icon='TRIA_RIGHT',
        )
        row = layout.row()
        row.label(text="Run.py Path")
        row = layout.row()
        row.prop(scene, "zpy_runpy_path")
        row = layout.row()
        row.operator(
            'scene.zpy_load_runpy',
            text='Pull',
            icon='IMPORT',
        )
        row.operator(
            'scene.zpy_push_runpy',
            text='Push',
            icon='EXPORT',
        )
        row = layout.row()
        row.label(text="Gin Config Path")
        row = layout.row()
        row.prop(scene, "zpy_gin_config_path")
        row = layout.row()
        row.operator(
            'scene.zpy_load_gin_config',
            text='Pull',
            icon='IMPORT',
        )
        row.operator(
            'scene.zpy_push_gin_config',
            text='Push',
            icon='EXPORT',
        )



"""
    Image version of Saver object.
"""
import logging
from datetime import date
from pathlib import Path
from typing import Dict, List, Tuple, Union

import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class ImageSaver(zpy.saver.Saver):
    """Holds the logic for saving image annotations at runtime."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.images = {}
        self.image_name_to_id = {}
        self.seg_annotations_color_to_id = {}

    @gin.configurable
    def add_image(self,
                  name: str = 'default image',
                  style: str = 'default',
                  output_path: Union[str, Path] = '/tmp/test.png',
                  frame: int = 0,
                  width: int = 0,
                  height: int = 0,
                  zero_indexed: bool = True,
                  **kwargs,
                  ) -> Dict:
        """ Add image to save object. """
        image = {
            'name': name,
            'style': style,
            'output_path': str(output_path),
            'relative_path': str(Path(output_path).relative_to(self.output_dir)),
            'frame': frame,
            'width': width,
            'height': height,
        }
        image.update(**kwargs)
        image['id'] = len(self.images.keys())
        image['id'] += 0 if zero_indexed else 1
        log.debug(f'Adding image: {zpy.files.pretty_print(image)}')
        self.images[image['id']] = image
        self.image_name_to_id[name] = image['id']
        return image

    @gin.configurable
    def add_annotation(self,
                       image: str = 'default image',
                       category: str = None,
                       subcategory: str = None,
                       subcategory_zero_indexed: bool = True,
                       seg_image: str = None,
                       seg_color:  Tuple[float] = None,
                       parse_on_add: bool = True,
                       **kwargs,
                       ) -> Dict:
        """ Add annotation. """
        image_id = self.image_name_to_id.get(image, None)
        assert image_id is not None, f'Could not find id for image {image}'
        assert category is not None, 'Must provide a category for annotation.'
        category_id = self.category_name_to_id.get(category, None)
        assert category_id is not None, f'Could not find id for category {category}'
        self.categories[category_id]['count'] += 1
        annotation = {
            'image_id': image_id,
            'category_id': category_id,
        }
        if subcategory is not None:
            subcategory_id = self.categories[category_id]['subcategories'].index(
                subcategory)
            self.categories[category_id]['subcategory_count'][subcategory_id] += 1
            subcategory_id += 0 if subcategory_zero_indexed else 1
            annotation['subcategory_id'] = subcategory_id
        annotation.update(**kwargs)
        annotation['id'] = len(self.annotations)
        log.info(f'Adding annotation: {zpy.files.pretty_print(annotation)}')
        # For segmentation images, add bbox/poly/mask annotation
        if seg_image is not None and seg_color is not None:
            seg_image_id = self.image_name_to_id.get(seg_image, None)
            assert seg_image_id is not None, f'Could not find id for image {seg_image}'
            annotation['seg_color'] = seg_color
            if self.seg_annotations_color_to_id.get(seg_image, None) is None:
                self.seg_annotations_color_to_id[seg_image] = {}
            self.seg_annotations_color_to_id[seg_image][seg_color] = annotation['id']
        self.annotations.append(annotation)
        # This call creates correspondences between segmentation images
        # and the annotations. It should be used after both the images
        # and annotations have been added to the saver.
        if parse_on_add:
            self.parse_annotations_from_seg_image(image_name=seg_image)
        return annotation

    def parse_annotations_from_seg_image(self,
                                         image_name: str,
                                         ) -> Dict:
        """ Populate annotation field based on segmentation image. """
        # Verify that file is segmentation image
        is_iseg = zpy.files.file_is_of_type(
            image_name, 'instance segmentation image')
        is_cseg = zpy.files.file_is_of_type(
            image_name, 'class segmentation image')
        if not (is_iseg or is_cseg):
            raise ValueError('Image is not segmentation image')
        seg_image_id = self.image_name_to_id.get(image_name, None)
        assert seg_image_id is not None, f'Could not find id for image {image_name}'
        image_path = self.images[seg_image_id]['output_path']
        if self.seg_annotations_color_to_id.get(image_name, None) is None:
            log.warning(f'No annotations found for {image_name}')
        for annotation in zpy.image.seg_to_annotations(image_path):
            if self.seg_annotations_color_to_id[image_name].get(annotation['color'], None) is None:
                log.warning(
                    f'No annotations found for color {annotation["color"]} in {image_name}')
                log.warning(
                    f'Available colors are {list(self.seg_annotations_color_to_id[image_name].keys())}')
                closest_color = zpy.color.closest_color(
                    annotation["color"],
                    list(self.seg_annotations_color_to_id[image_name].keys()))
                if closest_color is None:
                    log.warning(
                        f'Could not find close enough color, skipping ...')
                    continue
                else:
                    log.warning(f'Using closest color {closest_color}')
                    idx = self.seg_annotations_color_to_id[image_name][closest_color]
            else:
                idx = self.seg_annotations_color_to_id[image_name][annotation["color"]]
            self.annotations[idx].update(annotation)

    @gin.configurable
    def output_annotated_images(self,
                                num_annotated_images: int = 10,
                                ) -> None:
        """ Dump annotated sampled images to the meta folder. """
        log.info('Output annotated images...')
        import zpy.viz
        output_path = self.output_dir / self.HIDDEN_METAFOLDER_FILENAME
        output_path = zpy.files.verify_path(
            output_path, make=True, check_dir=True)
        for i, image in enumerate(self.images.values()):
            # Annotation images take a while
            if i >= num_annotated_images:
                return
            annotations = []
            for annotation in self.annotations:
                if annotation['image_id'] == image['id']:
                    annotations.append(annotation)
            if len(annotations) > 0:
                zpy.viz.draw_annotations(image_path=Path(image["output_path"]),
                                         annotations=annotations,
                                         categories=self.categories,
                                         output_path=output_path)

    @gin.configurable
    def output_meta_analysis(self,
                             image_sample_size: int = 50):
        """ Perform a full meta analysis.  """
        log.info(
            f'perform meta analysis image_sample_size:{image_sample_size}...')

        import zpy.files
        image_paths = [i['output_path']
                       for i in self.images.values() if i['style'] == 'default']
        image_paths = zpy.files.sample(
            image_paths, sample_size=image_sample_size)
        opened_images = [zpy.image.open_image(i) for i in image_paths]
        flat_images = zpy.image.flatten_images(opened_images)
        pixel_mean_std = zpy.image.pixel_mean_std(flat_images)

        meta_dict = {
            'number_images': len(self.images),
            'number_annotations': len(self.annotations),
            'number_categories': len(self.categories),
            'category_names': [c['name'] for c in self.categories.values()],
            'pixel_mean': np.array2string(pixel_mean_std["mean"], precision=2),
            'pixel_std': np.array2string(pixel_mean_std["std"], precision=2),
            'pixel_256_mean': np.array2string(pixel_mean_std["mean_256"], precision=0),
            'pixel_256_std': np.array2string(pixel_mean_std["std_256"], precision=0),
        }

        output_path = self.output_dir / self.HIDDEN_METAFOLDER_FILENAME
        output_path = zpy.files.verify_path(
            output_path, make=True, check_dir=True)
        self.write_datasheet(
            output_path / self.HIDDEN_DATASHEET_FILENAME, meta_dict)

        try:
            import zpy.viz
            zpy.viz.image_grid_plot(
                images=opened_images, output_path=output_path)
            zpy.viz.image_shape_plot(
                images=opened_images, output_path=output_path)
            zpy.viz.color_correlations_plot(
                flat_images=flat_images, output_path=output_path)
            zpy.viz.pixel_histograms(
                flat_images=flat_images, output_path=output_path)
            zpy.viz.category_barplot(
                categories=self.categories, output_path=output_path)
        except Exception as e:
            log.warning(f'Error when visualizing {e}')
            pass
